#include "MK20D5.h"
#include "DMA_K20.h"
 
 /**
  \fn          uint32_t DMA_UsedChannelCount ()
  \brief       Get count of used channels on DMA
*/ 
uint32_t DMA_UsedChannelCount () {
  uint32_t count = 0;
  uint32_t idx;
  
  for (idx = 0; idx < DMA_CHANNEL_COUNT; idx++) {
    count += (DMAMUX->CHCFG[idx] & DMAMUX_CHCFG_ENBL_MASK) ? 1 : 0; 
  }
  return count;
}

 /**
  \fn          void DMA_Initialize (uint32_t dma, uint32_t stream)
  \brief       Initialize DMA Stream
  \param[in]   stream  DMA number
  \param[in]   stream  DMA stream number
*/
void DMA_ChannelInitialize (uint32_t channel) {
  
  
  /* Initialize DMA clocks if they were not initialized yet */
  /* Enable clock to DMA module */
  SIM->SCGC7 |= (1U << SIM_SCGC7_DMA_SHIFT);
   
  /* Enable clock to DMAMUX module */
  SIM->SCGC6 |= (1U << SIM_SCGC6_DMAMUX_SHIFT);
   
  /* Configure DMA for normal operation */
  DMA0->CR = 0;  
    
  /* turn on interrupts */ 
  switch (channel) {
    case 0: {
      NVIC_ClearPendingIRQ(DMA0_IRQn);
      NVIC_EnableIRQ(DMA0_IRQn);
      break;
    }
    case 1: {
      NVIC_ClearPendingIRQ(DMA1_IRQn);
      NVIC_EnableIRQ(DMA1_IRQn);
      break;
    }
    case 2: {
      NVIC_ClearPendingIRQ(DMA2_IRQn);
      NVIC_EnableIRQ(DMA2_IRQn);
      break;
    }
    case 3: {
      NVIC_ClearPendingIRQ(DMA3_IRQn);
      NVIC_EnableIRQ(DMA3_IRQn);
      break;
    }
  }
  
}

/**
  \fn          void DMA_Uninitialize (uint32_t channel)
  \brief       Uninitialize DMA 
  \param[in]   DMA channel
*/
void DMA_ChannelUninitialize (uint32_t channel) {
  
  /* Disable channel */
  DMAMUX->CHCFG[channel] &= ~(DMAMUX_CHCFG_ENBL_MASK);
  
  
  /* Disable interrutp */
  switch (channel) {
    case 0: {
      NVIC_DisableIRQ(DMA0_IRQn);
      break;
    }
    case 1: {
      NVIC_DisableIRQ(DMA1_IRQn);
      break;
    }
    case 2: {
      NVIC_DisableIRQ(DMA2_IRQn);
      break;
    }
    case 3: {
      NVIC_DisableIRQ(DMA3_IRQn); 
      break;
    }
  }
  
  /* Turn DMA off if no channels are used at all - for energy saving */
  if (DMA_UsedChannelCount() == 0) {
    SIM->SCGC6 &= ~SIM_SCGC6_DMAMUX_MASK;
    SIM->SCGC7 &= ~SIM_SCGC7_DMA_MASK;
  }
     
}




 void DMA_ChannelConfigure (const DMA_INFO  *dma,
                                  uint32_t  source_addr,
                                  uint32_t  dest_addr,
                                  uint32_t  data_count,
                                  uint8_t   src_data_size,
                                  uint8_t   dest_data_size,
                                  uint16_t  src_offset,
                                  uint16_t  dest_offset,
                                  uint8_t   intr_enable){
                                          
                                            
  /* Set channel priority */                                           
  
    switch (dma->channel) {
    case 0: {
      DMA0->DCHPRI0 &= ~(DMA_DCHPRI0_CHPRI_MASK);
      DMA0->DCHPRI0 |= DMA_DCHPRI0_CHPRI(dma->priority);
      break;
    }
    case 1: {
      DMA0->DCHPRI1 &= ~(DMA_DCHPRI1_CHPRI_MASK);
      DMA0->DCHPRI1 |= DMA_DCHPRI1_CHPRI(dma->priority);
      break;
    }
    case 2: {
      DMA0->DCHPRI2 &= ~(DMA_DCHPRI2_CHPRI_MASK);
      DMA0->DCHPRI2 |= DMA_DCHPRI2_CHPRI(dma->priority);
      break;
    }
    case 3: {
      DMA0->DCHPRI3 &= ~(DMA_DCHPRI3_CHPRI_MASK);
      DMA0->DCHPRI3 |= DMA_DCHPRI3_CHPRI(dma->priority);
      break;
    }
    
  } 
  
  /* Enable interrupt for selected channel */
  DMA0->INT = (DMA_INT_INT0_MASK << dma->channel);
  
  
  /* Data source */
  DMA0->TCD[dma->channel].SADDR = source_addr;
  
  /* Data destination */
  DMA0->TCD[dma->channel].DADDR = dest_addr;
  
  /* Minor loop data count */
  DMA0->TCD[dma->channel].NBYTES_MLNO = DMA_SINGLE_TRANSFER;
  
  /* Offset at end of each write - source */
  DMA0->TCD[dma->channel].SOFF = src_offset;
  
  /* Offset at end of each write - destination */
  DMA0->TCD[dma->channel].DOFF = dest_offset;
  
  
  /* Source data size / size of sample */          
  DMA0->TCD[dma->channel].ATTR = DMA_ATTR_SMOD(DMA_ADDRESS_MOD_DISABLED) | DMA_ATTR_SSIZE(src_data_size) | 
                                 DMA_ATTR_DMOD(DMA_ADDRESS_MOD_DISABLED) | DMA_ATTR_DSIZE(dest_data_size);
                                 
  /* No source adjustment and no scatter and gathering at destination */                                 
  DMA0->TCD[dma->channel].SLAST = 0;
  DMA0->TCD[dma->channel].DLAST_SGA  = 0;
  
  
  /* Major loop iterration reset and set */ 
  DMA0->TCD[dma->channel].CITER_ELINKNO = 0;
  DMA0->TCD[dma->channel].BITER_ELINKNO = 0;
  DMA0->TCD[dma->channel].CITER_ELINKNO = data_count;
  
  /* Major loop interrupt enable */
  DMA0->TCD[dma->channel].CSR = DMA_CSR_INTMAJOR_MASK ;
  
  /* Disable HW requests when major loop finished */
  DMA0->TCD[dma->channel].CSR |= DMA_CSR_DREQ_MASK;  
  
  /* Disable channel to configure it */
  DMAMUX->CHCFG[dma->channel] = 0;
  
  /* Set data source */
  DMAMUX->CHCFG[dma->channel] = DMAMUX_CHCFG_SOURCE(dma->slot);
  
  
}




/// DMA Channel Events
__weak void DMA_Channel0_Event (uint32_t event) {};
__weak void DMA_Channel1_Event (uint32_t event) {};
__weak void DMA_Channel2_Event (uint32_t event) {};
__weak void DMA_Channel3_Event (uint32_t event) {};
  
       


/**
  \fn          void DMA_Channel0_IRQHandler (void)
  \brief       DMA Channel0 interrupt handler
*/
void DMA0_IRQHandler (void) {
  uint32_t events;

  events = (DMA0->TCD[0].CSR & DMA_STREAM_TRANSFER_COMPLETE);
  DMA0->CDNE &= ~(1 << 0);
  DMA_Channel0_Event(events);
}

/**
  \fn          void DMA_Channel1_IRQHandler (void)
  \brief       DMA Channel1 interrupt handler
*/
void DMA1_IRQHandler (void) {
  uint32_t events;

  events = (DMA0->TCD[1].CSR & DMA_STREAM_TRANSFER_COMPLETE);
  DMA0->CDNE &= ~(1 << 1);
  DMA_Channel1_Event(events);
}

/**
  \fn          void DMA_Channel2_IRQHandler (void)
  \brief       DMA Channel2 interrupt handler
*/
void DMA2_IRQHandler (void) {
  uint32_t events;

  events = (DMA0->TCD[2].CSR & DMA_STREAM_TRANSFER_COMPLETE);
  DMA0->CDNE &= ~(1 << 2);
  DMA_Channel2_Event(events);
}

/**
  \fn          void DMA_Channel3_IRQHandler (void)
  \brief       DMA Channel3 interrupt handler
*/
void DMA3_IRQHandler (void) {
  uint32_t events;

  events = (DMA0->TCD[3].CSR & DMA_STREAM_TRANSFER_COMPLETE);
  DMA0->CDNE &= ~(1 << 3);
  DMA_Channel3_Event(events);
}

