#ifndef __DMA_K20_H
#define __DMA_K20_H

#include <stdint.h>
#include <stdbool.h>
#include "MK20D5.h"

#if defined (__CC_ARM)
#define __FORCE_INLINE  static __forceinline
#else
#define __FORCE_INLINE  __STATIC_INLINE
#endif

#define _DMA_CHANNELx_EVENT(x)   DMA_Channel##x##_Event
#define DMA_CHANNELx_EVENT(x)    _DMA_CHANNELx_EVENT(x)

#define DMAMUX_SRC_UART0_TX     3
#define DMAMUX_SRC_UART1_TX     5
#define DMAMUX_SRC_UART2_TX     7

#define DMA_CHANNEL_COUNT     4

#define DMA_DATA_SIZE_8BIT     0
#define DMA_DATA_SIZE_16BIT    1
#define DMA_DATA_SIZE_32BIT    2
#define DMA_DATA_SIZE_16BYTE   4


#define DMA_SINGLE_TRANSFER 1

#define DMA_ADDRESS_MOD_DISABLED 0

#define DMA_STREAM_TRANSFER_COMPLETE    (1UL<<7)


#define DMA_MAJORITR_INT_ENABLE 1

// DMA Information definitions
typedef struct _DMA_INFO {
  uint8_t    channel;
  uint8_t    priority;
  uint8_t    slot;
} DMA_INFO;

/**
  \fn          void DMA_ChannelInitialize (uint32_t channel)
  \brief       Initialize DMA Channel
  \param[in]   DMA channel number
*/
extern void DMA_ChannelInitialize (uint32_t channel);

/**
  \fn          void DMA_ChannelUninitialize (uint32_t channel)
  \brief       Uninitialize DMA Channel
  \param[in]   DMA channel number
*/
extern void DMA_ChannelUninitialize (uint32_t channel);


/**
  \fn          void DMA_ChannelEnable (DMA_Stream_TypeDef *DMA_Stream)
  \brief       Enable channel and start memory to memory transfer
  \param[in]   channel  Channel number
*/
__FORCE_INLINE void DMA_ChannelEnable (uint32_t channel) {
    /* Enable channel */
  DMAMUX->CHCFG[channel] |= DMAMUX_CHCFG_ENBL_MASK;  
  /* Set Enable HW Request for channel */
  DMA0->ERQ = (DMA_ERQ_ERQ0_MASK << channel);
}

extern void DMA_Init_HARDCODED(uint32_t source_addr);
extern void DMA_Init_HARDCODED_MEMORY(uint32_t source_addr, uint32_t dest_addr);

/**
  \fn          bool DMA_ChannelStatus (uint32_t channel)
  \brief       Check if channel is enabled or disabled
  \param[in]   channel  DMA channel number
  \return      Stream Status
               - \b true Enabled
               - \b false Disabled
*/
__FORCE_INLINE bool DMA_ChannelStatus (uint32_t channel) {
  return (DMAMUX->CHCFG[channel] & DMAMUX_CHCFG_ENBL_MASK) ? true : false;
}

/**
  \fn          void DMA_StreamDisable (DMA_Stream_TypeDef *DMA_Stream)
  \brief       Disable stream and/or stop memory to memory transfer
  \param[in]   DMA_Stream  Pointer to DMA Stream peripheral
*/
__FORCE_INLINE void DMA_ChannelDisable (uint32_t channel) {
  DMA0->TCD[channel].CSR &= ~(DMA_CSR_START_MASK);
  DMAMUX->CHCFG[channel] &= ~(DMAMUX_CHCFG_ENBL_MASK);
}

/**
  \fn          void DMA_StreamConfigure (uint32_t   channel
                                         uint32_t   pr,
                                         uint8_t    slot);
  \brief       Configure DMA Stream for next transfer
  \param[in]   DMA_Stream specifies pointer to DMA Stream peripheral
  \param[in]   pr           Channel priority
  \param[in]   slot         Channel source slot
  \param[in]   source_addr  Data source address
  \param[in]   dest_addr    Data destination address
  \param[in]   data_size    Data count 


*/
extern void DMA_ChannelConfigure (const DMA_INFO  *dma,
                                          uint32_t  source_addr,
                                          uint32_t  dest_addr,
                                          uint32_t  data_count,
                                          uint8_t   src_data_size,
                                          uint8_t   dest_data_size,
                                          uint16_t  src_offset,
                                          uint16_t  dest_offset,
                                          uint8_t   intr_enable);



#endif /* __DMA_K20_H */

