#include "MK20D5.h"
#include "GPIO_K20.h"

GPIO_Type *GPIO_FromPort(PORT_Type *PORTx) {
  if      (PORTx == PORTA)
    return PTA;
  else if (PORTx == PORTB)
    return PTB;
  else if (PORTx == PORTC)
    return PTC;
  else if (PORTx == PORTD)
    return PTD;
  else if (PORTx == PORTE)
    return PTE;
  else 
    return 0;
}

/**
  \fn          void GPIO_PinWrite (GPIO_TypeDef *GPIOx, uint32_t num, uint32_t val)
  \brief       Write port pin
  \param[in]   GPIOx  Pointer to GPIO peripheral
  \param[in]   num    Port pin number
  \param[in]   val    Port pin value (0 or 1)
*/
void GPIO_PinWrite (PORT_Type *PORTx, uint32_t num, uint32_t val) {
  GPIO_Type *GPIOx;
  
  GPIOx = GPIO_FromPort(PORTx); 
  if (val & 1) {
    GPIOx->PSOR = (1UL << num);
  } else {
    GPIOx->PCOR = (1UL << num);
  }
}

/**
  \fn          uint32_t GPIO_PinRead (GPIO_TypeDef *GPIOx, uint32_t num)
  \brief       Read port pin
  \param[in]   GPIOx  Pointer to GPIO peripheral
  \param[in]   num    Port pin number
  \return      pin value (0 or 1)
*/
uint32_t GPIO_PinRead (PORT_Type *PORTx, uint32_t num) {
  GPIO_Type *GPIOx;
  
  GPIOx = GPIO_FromPort(PORTx); 
  return ((GPIOx->PDIR >> num) & 1);
}

/**
  \fn          void GPIO_PortWrite (GPIO_TypeDef *GPIOx, uint16_t mask, uint16_t val)
  \brief       Write port pins
  \param[in]   GPIOx  Pointer to GPIO peripheral
  \param[in]   mask   Selected pins
  \param[in]   val    Pin values
*/
void GPIO_PortWrite (PORT_Type *PORTx, uint16_t mask, uint16_t val) {
  GPIO_Type *GPIOx;
  
  GPIOx = GPIO_FromPort(PORTx); 
  GPIOx->PDOR = (GPIOx->PDOR & ~mask) | val;
}

/**
  \fn          uint16_t GPIO_PortRead (GPIO_TypeDef *GPIOx)
  \brief       Read port pins
  \param[in]   GPIOx  Pointer to GPIO peripheral
  \return      port pin inputs
*/
uint16_t GPIO_PortRead (PORT_Type *PORTx) {
  GPIO_Type *GPIOx;
  
  GPIOx = GPIO_FromPort(PORTx); 
  return (GPIOx->PDIR);
}


/**
  \fn          void GPIO_PortClock (GPIO_TypeDef *GPIOx, bool clock)
  \brief       Port Clock Control
  \param[in]   GPIOx  Pointer to GPIO peripheral
  \param[in]   clock  Enable or disable clock
*/
void GPIO_PortClock (PORT_Type *PORTx, bool clock) {

  if (clock) {
    if      (PORTx == PORTA) SIM->SCGC5 |=  (1U << SIM_SCGC5_PORTA_SHIFT);
    else if (PORTx == PORTB) SIM->SCGC5 |=  (1U << SIM_SCGC5_PORTB_SHIFT);
    else if (PORTx == PORTC) SIM->SCGC5 |=  (1U << SIM_SCGC5_PORTC_SHIFT);
    else if (PORTx == PORTD) SIM->SCGC5 |=  (1U << SIM_SCGC5_PORTD_SHIFT);
    else if (PORTx == PORTE) SIM->SCGC5 |=  (1U << SIM_SCGC5_PORTE_SHIFT);
  } else {
    if      (PORTx == PORTA) SIM->SCGC5 &= ~(1U << SIM_SCGC5_PORTA_SHIFT);
    else if (PORTx == PORTB) SIM->SCGC5 &= ~(1U << SIM_SCGC5_PORTB_SHIFT);
    else if (PORTx == PORTC) SIM->SCGC5 &= ~(1U << SIM_SCGC5_PORTC_SHIFT);
    else if (PORTx == PORTD) SIM->SCGC5 &= ~(1U << SIM_SCGC5_PORTD_SHIFT);
    else if (PORTx == PORTE) SIM->SCGC5 &= ~(1U << SIM_SCGC5_PORTE_SHIFT);
  }
} 



/**
  \fn          bool GPIO_PinAF (GPIO_TypeDef *GPIOx,
                                uint32_t      num,
                                uint32_t      af)
  \brief       Setup port pin alternate function
  \param[in]   GPIOx  Pointer to GPIO peripheral
  \param[in]   num    Port pin number
  \param[in]   af     \ref GPIO_AF_ID
  \return      true  - success
               false - error
*/

bool GPIO_PinAF (PORT_Type *PORTx,
                 uint32_t      num,
                 GPIO_AF_ID    af) {

  if (num >= 32) return false;

  PORTx->PCR[num] &= ~PORT_PCR_MUX_MASK;
  PORTx->PCR[num] |= af << PORT_PCR_MUX_SHIFT;

  return true;
}


