/* ----------------------------------------------------------------------
 * Copyright (C) 2013-2017 ARM Limited. All rights reserved.
 *  
 * $Date:        19. September 2017
 * $Revision:    V1.1
 *  
 * Driver:       Driver_UART0, Driver_UART1, Driver_UART2 
 *
 * Project:      UART Driver for Freescale Kinetis K20 
 * ---------------------------------------------------------------------- 
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 * 
 *   Configuration Setting               Value     UART Interface
 *   ---------------------               -----     -------------
 *   Connect to hardware via Driver_UART# = 0       use UART0
 *   Connect to hardware via Driver_UART# = 1       use UART1
 *   Connect to hardware via Driver_UART# = 2       use UART2
 * -------------------------------------------------------------------- */

/* History:
 *  Version 1.1
 *    Added support for CMSIS-RTOS2
 *  Version 0.01
 *    Initial release
 */

#include "RTE_Components.h"

#if       defined(RTE_CMSIS_RTOS2)
#include "cmsis_os2.h"
#elif     defined(RTE_CMSIS_RTOS)
#include "cmsis_os.h"
#endif

#include "MK20D5.h"

#include "GPIO_K20.h"
#include "DMA_K20.h"

#include "Driver_UART.h"


#include "RTE_Device.h"
#include "RTE_Components.h"

#include <string.h>




#define UART_DRV_VERSION ARM_DRV_VERSION_MAJOR_MINOR(1,1)   /* driver version */

#if ((defined(RTE_Drivers_UART0) || \
      defined(RTE_Drivers_UART1) || \
      defined(RTE_Drivers_UART2))   \
     && !RTE_UART0                  \
     && !RTE_UART1                  \
     && !RTE_UART2)                 
#error "UART not configured!"
#endif

#if (RTE_UART0 && RTE_UART0_RX_DMA)
#warning "UART0 does not support receive DMA!"
#endif

#if (RTE_UART1 && RTE_UART1_RX_DMA)
#warning "UART1 does not support receive DMA!"
#endif

#if (RTE_UART2 && RTE_UART2_RX_DMA)
#warning "UART2 does not support receive DMA!"
#endif


#define DMA_TX  ((RTE_UART0 && RTE_UART0_TX_DMA) ||       \
                 (RTE_UART1 && RTE_UART1_TX_DMA) ||       \
                 (RTE_UART2 && RTE_UART2_TX_DMA)) 


/* Driver Version */
static const ARM_DRV_VERSION DriverVersion = {
  ARM_UART_API_VERSION,
  UART_DRV_VERSION
};

/* UART Data Information (Run-Time) */
typedef struct {
  volatile uint16_t  idx_in;
  volatile uint16_t  idx_out;
  volatile bool      empty;
           uint16_t  threshold;
           uint16_t  transfer_len;
} UART_DATA_INFO;

/* UART Information (Run-Time) */
typedef struct {
  bool               power;
  ARM_UART_SignalEvent_t cb_event;
  uint32_t           event_mask;
#if   defined(RTE_CMSIS_RTOS2)
  osSemaphoreId_t    semaphore;
  osMutexId_t        mutex_write;
  osMutexId_t        mutex_read;
#elif defined(RTE_CMSIS_RTOS)
  osSemaphoreId      semaphore;
  osMutexId          mutex_write;
  osMutexId          mutex_read;
#endif
  UART_DATA_INFO     rx;
  UART_DATA_INFO     tx;
  bool               com_break;
  bool               rts_cts_flow;
} UART_INFO;

/* UART Resources definition */
typedef struct {
  const ARM_UART_CAPABILITIES  *capabilities;       // Capabilities
        PORT_PIN_ID         pin_tx;             // TX  Pin identifier
        PORT_PIN_ID         pin_rx;             // RX  Pin identifier
        PORT_PIN_ID         pin_cts;            // CTS Pin identifier
        PORT_PIN_ID         pin_rts;            // RTS Pin identifier
        UART_Type          *uart;               // Pointer to UART peripheral
  //enum  IRQn                uart_lon_irq;       // UART LON IRQ Number
  enum  IRQn                uart_rx_tx_irq;     // UART RX TX IRQ Number
  enum  IRQn                uart_err_irq;       // UART ERROR IRQ Number
  __IO  uint32_t           *clock_reg;          // Pointer to Clock register
        uint32_t            clock_mask;         // Clock register mask
  __IO  uint32_t           *clock_src_reg;      // Pointer to clock source register
        uint32_t            clock_src_reg_mask; // Clock source register 
        uint32_t            clock;              // Peripheral clock
        bool                hw_flow;            // Hardware flow control
        bool                manual_rts_cts;     // Manual RTS CTS control
        uint8_t            *tx_buffer;          // Transmit buffer
        uint32_t            tx_buffer_size;     // Transmit buffer size
        uint8_t            *rx_buffer;          // Receive buffer
        uint32_t            rx_buffer_size;     // Receive buffer size
        uint32_t            flow_ctrl_threshold;// Flow control threshold
  const DMA_INFO           *dma_tx;             // Transmit DMA configurationn
#if     defined(RTE_CMSIS_RTOS)
  const osSemaphoreDef_t   *semaphore;          // Tx Semaphore definition
  const osMutexDef_t       *mutex_write;        // Write Mutex definition
  const osMutexDef_t       *mutex_read;         // Write Mutex definition
#endif
        UART_INFO          *info;               // Run-Time information
} const UART_RESOURCES;


/* UART0 */
#if (RTE_UART0 != 0)

#ifndef __UART0_TX_BUFFER_SIZE
#define __UART0_TX_BUFFER_SIZE         128  // Buffer size must be 2^n
#endif

#ifndef __UART0_RX_BUFFER_SIZE
#define __UART0_RX_BUFFER_SIZE         128  // Buffer size must be 2^n
#endif

#ifndef __UART0_FLOW_CONTROL_THRESHOLD
#define __UART0_FLOW_CONTROL_THRESHOLD  16  // Buffer size must be 2^n
#endif


/* UART0 Capabilities */
static const UART_CAPABILITIES UART0_Capabilities = {
  __UART0_TX_BUFFER_SIZE,           // Transmit Buffer Size
  __UART0_RX_BUFFER_SIZE,           // Receive Buffer Size
  0,                                // Receive character timouet event
  0,
  0,
  0,                                // DSR Line available
  0,                                // Signal DSR change event
  0,                                // DCD Line available
  0,                                // Signal DCD change event
  0,                                // RI Line available
  0,                                // Signal RI change event
  0
};

/* UART0 DMA */
#if (RTE_UART0_TX_DMA != 0)
static const DMA_INFO UART0_DMA_TX = {
  RTE_UART0_TX_DMA_CHANNEL,
  RTE_UART0_TX_DMA_PRIORITY,
  DMAMUX_SRC_UART0_TX
};
#endif

#if defined(RTE_CMSIS_RTOS)
/* UART0 Semaphores */
osSemaphoreDef(UART0_Semaphore);

/* UART0 Mutex */
osMutexDef(UART0_Mutex_write);
osMutexDef(UART0_Mutex_read);
#endif


/* UART0 Data Info */
// UART0 Buffer size must be 2^n
struct {
  uint8_t  tx[__UART0_TX_BUFFER_SIZE];
  uint8_t  rx[__UART0_RX_BUFFER_SIZE];
} UART0_Buffer;

static UART_INFO UART0_Info;

/* UART0 Resources */
static UART_RESOURCES UART0_Resources = {
  &UART0_Capabilities,
  { RTE_UART0_TX_PORT,  RTE_UART0_TX_BIT,  RTE_UART0_TX_AF },   // tx pin
  { RTE_UART0_RX_PORT,  RTE_UART0_RX_BIT,  RTE_UART0_RX_AF },   // rx pin
  { RTE_UART0_CTS_PORT, RTE_UART0_CTS_BIT, RTE_UART0_CTS_AF},   // cts pin
  { RTE_UART0_RTS_PORT, RTE_UART0_RTS_BIT, RTE_UART0_RTS_AF},   // rts pin
  UART0,                                                        // UART peripheral
  UART0_LON_IRQn,                                               // UART LON IRQ number
  UART0_RX_TX_IRQn,                                             // UART RX TX IRQ number
  UART0_ERR_IRQn,                                               // UART ERR IRQ number
  &SIM->SCGC4,                                                  // system clock gating
  SIM_SCGC4_UART0_MASK,                                         // system clock mask
  &SIM->SOPT5,                                                  // system option register (source for RX and TX lines) 
  SIM_SOPT5_UART0RXSRC_MASK | SIM_SOPT5_UART0TXSRC_MASK,        // system option register mask (source for RX and TX lines - both used)
  RTE_SYSCLK,                                                   // peripheral clock - system clock
  RTE_UART0_HW_FLOW,                                            // HW flow control
  RTE_UART0_MANUAL_FLOW,                                        // manual RTS/CTS control
  UART0_Buffer.tx,                                              // transmit buffer     
  __UART0_TX_BUFFER_SIZE,                                       // transmit buffer size
  UART0_Buffer.rx,                                              // receive buffer 
  __UART0_RX_BUFFER_SIZE,                                       // receive buffer size
  __UART0_FLOW_CONTROL_THRESHOLD,                               // flow control threshold
#if (RTE_UART0_TX_DMA != 0)
  &UART0_DMA_TX,
#else
  NULL,
#endif
#if defined(RTE_CMSIS_RTOS)
  osSemaphore(UART0_Semaphore),
  osMutex(UART0_Mutex_write),
  osMutex(UART0_Mutex_read),
#endif
  &UART0_Info
};
#endif



/* UART1 */
#if (RTE_UART1 != 0)

#ifndef __UART1_TX_BUFFER_SIZE
#define __UART1_TX_BUFFER_SIZE         128  // Buffer size must be 2^n
#endif

#ifndef __UART1_RX_BUFFER_SIZE
#define __UART1_RX_BUFFER_SIZE         128  // Buffer size must be 2^n
#endif

#ifndef __UART1_FLOW_CONTROL_THRESHOLD
#define __UART1_FLOW_CONTROL_THRESHOLD  16  // Buffer size must be 2^n
#endif


/* UART1 Capabilities */
static const ARM_UART_CAPABILITIES UART1_Capabilities = {
  __UART1_TX_BUFFER_SIZE,           // Transmit Buffer Size
  __UART1_RX_BUFFER_SIZE,           // Receive Buffer Size
  0,                                // Receive character timouet event
  0,
  0,
  0,                                // DSR Line available
  0,                                // Signal DSR change event
  0,                                // DCD Line available
  0,                                // Signal DCD change event
  0,                                // RI Line available
  0,                                // Signal RI change event
  0
};

/* UART1 DMA */
#if (RTE_UART1_TX_DMA != 0)
static const DMA_INFO UART1_DMA_TX = {
  RTE_UART1_TX_DMA_CHANNEL,
  RTE_UART1_TX_DMA_PRIORITY,
  DMAMUX_SRC_UART1_TX
};
#endif

#if defined(RTE_CMSIS_RTOS)
/* UART1 Semaphores */
osSemaphoreDef(UART1_Semaphore);

/* UART1 Mutex */
osMutexDef(UART1_Mutex_write);
osMutexDef(UART1_Mutex_read);
#endif


/* UART1 Data Info */
// UART1 Buffer size must be 2^n
struct {
  uint8_t  tx[__UART1_TX_BUFFER_SIZE];
  uint8_t  rx[__UART1_RX_BUFFER_SIZE];
} UART1_Buffer;

static UART_INFO UART1_Info;

/* UART1 Resources */
static UART_RESOURCES UART1_Resources = {
  &UART1_Capabilities,
  { RTE_UART1_TX_PORT,  RTE_UART1_TX_BIT,  RTE_UART1_TX_AF },   // tx pin
  { RTE_UART1_RX_PORT,  RTE_UART1_RX_BIT,  RTE_UART1_RX_AF },   // rx pin
  { RTE_UART1_CTS_PORT, RTE_UART1_CTS_BIT, RTE_UART1_CTS_AF},   // cts pin
  { RTE_UART1_RTS_PORT, RTE_UART1_RTS_BIT, RTE_UART1_RTS_AF},   // rts pin 
  UART1,                                                        // UART peripheral
  //0,                                                      // UART LON IRQ number
  UART1_RX_TX_IRQn,                                             // UART RX TX IRQ number
  UART1_ERR_IRQn,                                               // UART ERR IRQ number
  &SIM->SCGC4,                                                  // system clock gating
  SIM_SCGC4_UART1_MASK,                                         // system clock mask
  &SIM->SOPT5,                                                  // system option register (source for RX and TX lines) 
  SIM_SOPT5_UART1RXSRC_MASK | SIM_SOPT5_UART1TXSRC_MASK,        // system option register mask (source for RX and TX lines - both used)
  RTE_SYSCLK,                                                   // peripheral clock - system clock
  RTE_UART1_HW_FLOW,                                            // HW flow control
  RTE_UART1_MANUAL_FLOW,                                        // manual RTS/CTS control
  UART1_Buffer.tx,                                              // transmit buffer     
  __UART1_TX_BUFFER_SIZE,                                       // transmit buffer size
  UART1_Buffer.rx,                                              // receive buffer 
  __UART1_RX_BUFFER_SIZE,                                       // receive buffer size
  __UART1_FLOW_CONTROL_THRESHOLD,                               // flow control threshold
#if (RTE_UART1_TX_DMA != 0)
  &UART1_DMA_TX,
#else
  NULL,
#endif
#if defined(RTE_CMSIS_RTOS)
  osSemaphore(UART1_Semaphore),
  osMutex(UART1_Mutex_write),
  osMutex(UART1_Mutex_read),
#endif
  &UART1_Info
};
#endif


/* UART2 */
#if (RTE_UART2 != 0)

#ifndef __UART2_TX_BUFFER_SIZE
#define __UART2_TX_BUFFER_SIZE         128  // Buffer size must be 2^n
#endif

#ifndef __UART2_RX_BUFFER_SIZE
#define __UART2_RX_BUFFER_SIZE         128  // Buffer size must be 2^n
#endif

#ifndef __UART2_FLOW_CONTROL_THRESHOLD
#define __UART2_FLOW_CONTROL_THRESHOLD  16  // Buffer size must be 2^n
#endif


/* UART2 Capabilities */
static const UART_CAPABILITIES UART2_Capabilities = {
  __UART2_TX_BUFFER_SIZE,           // Transmit Buffer Size
  __UART2_RX_BUFFER_SIZE,           // Receive Buffer Size
  0,                                // Receive character timouet event
  0,
  0,
  0,                                // DSR Line available
  0,                                // Signal DSR change event
  0,                                // DCD Line available
  0,                                // Signal DCD change event
  0,                                // RI Line available
  0,                                // Signal RI change event
  0
};

/* UART2 DMA */
#if (RTE_UART_TX_DMA != 0)
static const DMA_INFO UART_DMA_TX = {
  RTE_UART_TX_DMA_CHANNEL,
  RTE_UART_TX_DMA_PRIORITY,
  DMAMUX_SRC_UART2_TX
};
#endif

#if defined(RTE_CMSIS_RTOS)
/* UART2 Semaphores */
osSemaphoreDef(UART2_Semaphore);

/* UART2 Mutex */
osMutexDef(UART2_Mutex_write);
osMutexDef(UART2_Mutex_read);
#endif


/* UART2 Data Info */
// UART2 Buffer size must be 2^n
struct {
  uint8_t  tx[__UART2_TX_BUFFER_SIZE];
  uint8_t  rx[__UART2_RX_BUFFER_SIZE];
} UART2_Buffer;

static UART_INFO UART2_Info;

/* UART0 Resources */
static UART_RESOURCES UART2_Resources = {
  &UART2_Capabilities, 
  { RTE_UART2_TX_PORT,  RTE_UART2_TX_BIT,  RTE_UART2_TX_AF },   // tx pin
  { RTE_UART2_RX_PORT,  RTE_UART2_RX_BIT,  RTE_UART2_RX_AF },   // rx pin
  { RTE_UART2_CTS_PORT, RTE_UART2_CTS_BIT, RTE_UART2_CTS_AF},   // cts pin
  { RTE_UART2_RTS_PORT, RTE_UART2_RTS_BIT, RTE_UART2_RTS_AF},   // rts pin
  UART2,                                                        // UART peripheral
  0,                                                            // UART LON IRQ number
  UART2_RX_TX_IRQn,                                             // UART RX TX IRQ number
  UART2_ERR_IRQn,                                               // UART ERR IRQ number
  &SIM->SCGC4,                                                  // system clock gating
  SIM_SCGC4_UART2_MASK,                                         // system clock mask
  NULL,                                                         // system option register (source for RX and TX lines) 
  0,                                                            // system option register mask (source for RX and TX lines - both used)
  RTE_BUSCLK,                                                   // peripheral clock - system clock
  RTE_UART2_HW_FLOW,                                            // HW flow control
  RTE_UART2_MANUAL_FLOW,                                        // manual RTS/CTS control
  UART2_Buffer.tx,                                              // transmit buffer     
  __UART2_TX_BUFFER_SIZE,                                       // transmit buffer size
  UART2_Buffer.rx,                                              // receive buffer 
  __UART2_RX_BUFFER_SIZE,                                       // receive buffer size
  __UART2_FLOW_CONTROL_THRESHOLD,                               // flow control threshold
#if (RTE_UART2_TX_DMA != 0)
  &UART2_DMA_TX,
#else
  NULL,
#endif
#if defined(RTE_CMSIS_RTOS)
  osSemaphore(UART2_Semaphore),
  osMutex(UART2_Mutex_write),
  osMutex(UART2_Mutex_read),
#endif
  &UART2_Info
};
#endif


/**
  \fn          DRV_VERSION UARTx_GetVersion (void)
  \brief       Get driver version.
  \return      \ref DRV_VERSION
*/
static ARM_DRV_VERSION UARTX_GetVersion (void) {
  return DriverVersion;
}


/**
  \fn          UART_CAPABILITIES UART_GetCapabilities (UART_RESOURCES *ptr_uart)
  \brief       Get driver capabilities.
  \param[in]   ptr_uart    Pointer to UART resources
  \return      \ref UART_CAPABILITIES
*/
static ARM_UART_CAPABILITIES UART_GetCapabilities (UART_RESOURCES *ptr_uart) {
  return *ptr_uart->capabilities;
}

#if (RTE_UART0)
  static UART_CAPABILITIES UART0_GetCapabilities (void) {
    return UART_GetCapabilities (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_CAPABILITIES UART1_GetCapabilities (void) {
    return UART_GetCapabilities (&UART1_Resources);
  }
#endif
	
#if (RTE_UART2)
  static UART_CAPABILITIES UART2_GetCapabilities (void) {
    return UART_GetCapabilities (&UART2_Resources);
  }
#endif	




/**
  \fn          ARM_UART_STATUS UART_Initialize (UART_SignalEvent_t cb_event,
                                            uint32_t           event_mask
                                            UART_RESOURCES     *ptr_uart)
  \brief       Initialize UART Interface.
  \param[in]   cb_event    Pointer to \ref UART_SignalEvent
  \param[in]   event_mask
  \param[in]   ptr_uart    Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_Initialize (ARM_UART_SignalEvent_t cb_event,
                                    uint32_t           event_mask,
                                    UART_RESOURCES     *ptr_uart) {
  bool status;
                                        
                                                                                                          
                                        
  /* Initialize UART Run-Time Resources */
  ptr_uart->info->power           = false;  
  ptr_uart->info->cb_event        = cb_event;
  ptr_uart->info->event_mask      = (cb_event == NULL) ? event_mask : 0;
  
  ptr_uart->info->tx.idx_in       = 0;
  ptr_uart->info->tx.idx_out      = 0;
  ptr_uart->info->tx.empty        = true;
  ptr_uart->info->tx.transfer_len = 0;
  ptr_uart->info->rx.idx_in       = 0;
  ptr_uart->info->rx.idx_out      = 0;
  ptr_uart->info->rx.empty        = true;
  ptr_uart->info->rx.transfer_len = 0;
  ptr_uart->info->com_break       = false;

  if (ptr_uart->info->semaphore   == NULL) {
#if   defined(RTE_CMSIS_RTOS2)
    ptr_uart->info->semaphore     = osSemaphoreNew(1, 1, NULL);
#elif defined(RTE_CMSIS_RTOS)
    ptr_uart->info->semaphore     = osSemaphoreCreate(ptr_uart->semaphore, 1);
#endif
  }
  if (ptr_uart->info->mutex_write == NULL) {
#if   defined(RTE_CMSIS_RTOS2)
    ptr_uart->info->mutex_write   = osMutexNew(NULL);
#elif defined(RTE_CMSIS_RTOS)
    ptr_uart->info->mutex_write   = osMutexCreate(ptr_uart->mutex_write);
#endif
  }
  if (ptr_uart->info->mutex_read  == NULL) {
#if   defined(RTE_CMSIS_RTOS2)
    ptr_uart->info->mutex_read    = osMutexNew(NULL);
#elif defined(RTE_CMSIS_RTOS)
    ptr_uart->info->mutex_read    = osMutexCreate(ptr_uart->mutex_read);
#endif
  }

  if (ptr_uart->info->semaphore   == NULL) return ARM_UART_ERROR;
  if (ptr_uart->info->mutex_write == NULL) return ARM_UART_ERROR;
  if (ptr_uart->info->mutex_read  == NULL) return ARM_UART_ERROR; 
  
  /* Initialize DMA Streams */
  #if DMA_TX
    if (ptr_uart->dma_tx) {
      DMA_ChannelInitialize(ptr_uart->dma_tx->channel);
   }
  #endif
  
  
  // configure TX pin
  GPIO_PortClock       (ptr_uart->pin_tx.port, true);
  status = GPIO_PinAF  (ptr_uart->pin_tx.port, 
                        ptr_uart->pin_tx.num, 
                        ptr_uart->pin_tx.af);
  if (!status) return ARM_UART_ERROR;

  /* Configure RX pin */
 GPIO_PortClock       (ptr_uart->pin_rx.port, true);
 status = GPIO_PinAF  (ptr_uart->pin_rx.port, 
                       ptr_uart->pin_rx.num, 
                       ptr_uart->pin_rx.af);
  if (!status) return ARM_UART_ERROR;

  /* select source for RX and TX lines (UART2 has no assigned clock src register) */
  if (*ptr_uart->clock_src_reg != NULL)
    *ptr_uart->clock_src_reg &= ~ptr_uart->clock_src_reg_mask;

  /* Enable UART interrupts */
  //if (ptr_uart->uart_lon_irq != 0){
    /* only uart0 */
  //  NVIC_ClearPendingIRQ (ptr_uart->uart_lon_irq);
  //  NVIC_EnableIRQ (ptr_uart->uart_lon_irq);
  // }
   
  NVIC_ClearPendingIRQ (ptr_uart->uart_rx_tx_irq);
  NVIC_EnableIRQ (ptr_uart->uart_rx_tx_irq);
  
  NVIC_ClearPendingIRQ (ptr_uart->uart_err_irq);
  NVIC_EnableIRQ (ptr_uart->uart_err_irq);
  
  return ARM_UART_OK;
} 

#if (RTE_UART0 != 0)
 static ARM_UART_STATUS UART0_Initialize (UART_SignalEvent_t cb_event,
                                      uint32_t           event_mask) {
   return UART_Initialize(cb_event, event_mask, &UART0_Resources);
  }
#endif
	
#if (RTE_UART1 != 0)
 static ARM_UART_STATUS UART1_Initialize (ARM_UART_SignalEvent_t cb_event,
                                      uint32_t           event_mask) {
   return  UART_Initialize(cb_event, event_mask, &UART1_Resources);
  }
#endif	
	
#if (RTE_UART2 != 0)
 static ARM_UART_STATUS UART2_Initialize (UART_SignalEvent_t cb_event,
                                      uint32_t           event_mask) {
   return UART_Initialize(cb_event, event_mask, &UART2_Resources);
  }
#endif



/**
  \fn          ARM_UART_STATUS UART_Uninitialize (UART_RESOURCES *ptr_uart)
  \brief       De-initialize UART Interface.
  \param[in]   ptr_uart  Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_Uninitialize (UART_RESOURCES *ptr_uart) {
  bool status;

  /* Enable UART clock */
  *ptr_uart->clock_reg |=  ptr_uart->clock_mask;
  
  // Reset UART peripheral
  //*ptr_uart->reset_reg |=  ptr_uart->reset_mask;
  //__NOP(); __NOP(); __NOP(); __NOP();
  //*ptr_uart->reset_reg &= ~ptr_uart->reset_mask;

  /* Disable UART clock */
  *ptr_uart->clock_reg &= ~ptr_uart->clock_mask;


  /* Clear and Disable UART IRQ */
  //if (ptr_uart->uart_lon_irq != 0) {
  //  NVIC_ClearPendingIRQ(ptr_uart->uart_lon_irq);
  //  NVIC_DisableIRQ(ptr_uart->uart_lon_irq);
  //}

  NVIC_ClearPendingIRQ(ptr_uart->uart_rx_tx_irq);
  NVIC_DisableIRQ(ptr_uart->uart_rx_tx_irq);
  
  NVIC_ClearPendingIRQ(ptr_uart->uart_err_irq);
  NVIC_DisableIRQ(ptr_uart->uart_err_irq);


  /* Uninitialize UART Run-Time Resources */
  osSemaphoreDelete(ptr_uart->info->semaphore);
  osMutexDelete(ptr_uart->info->mutex_write);
  osMutexDelete(ptr_uart->info->mutex_read);

  /* Uninitialize DMA Streams */
  #if DMA_TX
  if (ptr_uart->dma_tx) {
    DMA_ChannelUninitialize(ptr_uart->dma_tx->channel);
  }
  #endif 
  
  /* Unconfigure TX pin */
  status = GPIO_PinAF         (ptr_uart->pin_tx.port,
                               ptr_uart->pin_tx.num,
                               GPIO_AF_GPIO);
  if (!status) return ARM_UART_ERROR;
  
  /* Unconfigure RX pin */
  status = GPIO_PinAF         (ptr_uart->pin_rx.port,
                               ptr_uart->pin_rx.num,
                               GPIO_AF_GPIO);
  if (!status) return ARM_UART_ERROR;
  

  if (ptr_uart->hw_flow == true) {
    /* Unconfigure CTS pin */
    status = GPIO_PinAF       (ptr_uart->pin_cts.port,
                               ptr_uart->pin_cts.num,
                               GPIO_AF_GPIO);
    if (!status) return ARM_UART_ERROR;
  }

  if (ptr_uart->hw_flow == true) {
    /* Unconfigure RTS pin */
    status = GPIO_PinAF       (ptr_uart->pin_rts.port,
                               ptr_uart->pin_rts.num,
                               GPIO_AF_GPIO);
    if (!status) return ARM_UART_ERROR;
  }

  return ARM_UART_OK;
}

#if (RTE_UART0 != 0)
  static ARM_UART_STATUS UART0_Uninitialize (void) {
    return UART_Uninitialize (&UART0_Resources);
  }
#endif

#if (RTE_UART1 != 0)
  static ARM_UART_STATUS UART1_Uninitialize (void) {
    return UART_Uninitialize (&UART1_Resources);
  }
#endif

#if (RTE_UART2 != 0)
  static ARM_UART_STATUS UART2_Uninitialize (void) {
    return UART_Uninitialize (&UART2_Resources);
  }
#endif

/**
  \fn          ARM_UART_STATUS UART_PowerControl (POWER_STATE     state,
                                              UART_RESOURCES *ptr_uart)
  \brief       Controls UART Interface Power.
  \param[in]   state    Power state
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_PowerControl (ARM_POWER_STATE state, UART_RESOURCES *ptr_uart) {	
  uint8_t error_intr = 0;
  uint8_t traffic_intr = 0;

  if (state == ARM_POWER_OFF) {
    /* Set power propery */
    ptr_uart->info->power = false;

    /* Enable UART Clock */
    *ptr_uart->clock_reg |= ptr_uart->clock_mask;

    /* reset parity */
    ptr_uart->uart->C1 = 0;
    
    /* disable transmitter and receiver / DMA */
    ptr_uart->uart->C2 = 0;
    
    /* disable interrupts */
    ptr_uart->uart->C3 = 0;

    /* Disable UART Clock */
    *ptr_uart->clock_reg &= ~ptr_uart->clock_mask;

    /* Disable DMA Streams */
    if (ptr_uart->dma_tx) {
      
      /* Disable DMA for TX */
      ptr_uart->uart->C5 &= ~(UART_C5_TDMAS_MASK);
      
      /* Disable DMA channel */
      DMA_ChannelDisable(ptr_uart->dma_tx->channel);
    }

  } else if (state == ARM_POWER_FULL) {
    
    /* Set power propery */
    ptr_uart->info->power = true;
    
    /* Enable UART Clock */
    *ptr_uart->clock_reg |= ptr_uart->clock_mask;
    
    // Clear status register
    
    // lin mode enable
    //ptr_uart->uart->CR2   |= USART_CR2_LINEN;

    // line break detection

    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_BREAK))
      ptr_uart->uart->BDH |= (1U << UART_BDH_LBKDIE_SHIFT);

    // idle line interrupt enable
    //ptr_uart->uart->C2 |= (1U << UART_C2_ILIE_SHIFT);
    //traffic_intr |= UART_C2_ILIE_MASK;                 

    /* Transmission complete interrupt enable */
    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_TX_DONE))
      traffic_intr |= UART_C2_TCIE_MASK;

    /* Overrun error interrupt enable */
    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_OVERRUN))
      error_intr |= UART_C3_ORIE_MASK;

    /* Noise error interrupt enable */
    //ptr_uart->uart->C3 |= (1U << UART_C3_NEIE_SHIFT);
    //  error_intr |= UART_C3_NEIE_MASK;

    /* Framing error interrupt enable */
    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_FRAMING_ERROR))
      error_intr |= UART_C3_FEIE_MASK;

    /* Parity error interrupt enable */
    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_PARITY_ERROR))
      error_intr |= UART_C3_PEIE_MASK;
    
    /* Enable UART / Receiver and transmitter */
    traffic_intr |= (UART_C2_RE_MASK | UART_C2_TE_MASK);

    /* Enable interrupts for RX and TX */
    traffic_intr |= (UART_C2_RIE_MASK | UART_C2_TIE_MASK);
    
    
    /* Enable DMA */
    #if DMA_TX 
      /* Disable transfer complete interrupt */
      traffic_intr &= ~(UART_C2_TCIE_MASK);    
    
      /* Enable interrupt requests for DMA */
      traffic_intr |= UART_C2_TIE_MASK;
      
      /* Enable DMA for TX */
      ptr_uart->uart->C5 |= UART_C5_TDMAS_MASK;  
    
      /* Initialize DMA channel */
      DMA_ChannelInitialize(ptr_uart->dma_tx->channel);
    #endif
   
    /* Set registers C3 and C2 */
    ptr_uart->uart->C3 = error_intr;
    ptr_uart->uart->C2 = traffic_intr;
    

    if ((ptr_uart->hw_flow == true) && (ptr_uart->manual_rts_cts == false)) {
      GPIO_PinWrite(ptr_uart->pin_rts.port, ptr_uart->pin_rts.num, 0);
    }

  }

  return ARM_UART_OK;
}

#if (RTE_UART0 != 0)
  static ARM_UART_STATUS UART0_PowerControl (POWER_STATE state) {
    return UART_PowerControl (state, &UART0_Resources);
  }
#endif

#if (RTE_UART1 != 0)
  static ARM_UART_STATUS UART1_PowerControl (ARM_POWER_STATE state) {
    return UART_PowerControl (state, &UART1_Resources);
  }
#endif

#if (RTE_UART2 != 0)
  static ARM_UART_STATUS UART2_PowerControl (POWER_STATE state) {
    return UART_PowerControl (state, &UART2_Resources);
  }
#endif

/**
  \fn          ARM_UART_STATUS UART_Configure (uint32_t           baudrate,
                                           uint8_t            data_bits,
                                           UART_PARITY        parity,
                                           UART_STOP_BITS     stop_bits,
                                           UART_FLOW_CONTROL  flow_control,
                                           UART_RESOURCES    *ptr_uart)
  \brief       Configure UART Interface.
  \param[in]   baudrate      Requested baudrate in bits/s
  \param[in]   data_bits     Number of data bits
  \param[in]   parity        \ref UART_PARITY
  \param[in]   stop_bits     \ref UART_STOP_BITS
  \param[in]   flow_control  \ref UART_FLOW_CONTROL
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/  
static ARM_UART_STATUS UART_Configure (uint32_t           baudrate,
                                   uint8_t            data_bits,
                                   ARM_UART_PARITY        parity,
                                   ARM_UART_STOP_BITS     stop_bits,
                                   ARM_UART_FLOW_CONTROL  flow_control,
                                   UART_RESOURCES    *ptr_uart) {
  uint32_t brCalc;
  uint32_t brDivider, brFract;
  uint32_t rts_pin = 1;

  if (ptr_uart->info->power == false) { return ARM_UART_ERROR; }
 
  if ((ptr_uart->hw_flow == true) && (ptr_uart->manual_rts_cts == false)) {
    rts_pin = GPIO_PinRead(ptr_uart->pin_rts.port, ptr_uart->pin_rts.num);
   if (rts_pin == 0)
      GPIO_PinWrite(ptr_uart->pin_rts.port, ptr_uart->pin_rts.num, 1);
  }

  // Disable UART transmiter and receiver
  ptr_uart->uart->C2 &= ~(UART_C2_RE_MASK | UART_C2_TE_MASK);

  // Baudrate
  brDivider = (ptr_uart->clock / 16 / baudrate);  // calculate baud rate divider from desired clock
  brFract   = (uint8_t)((float)((((float)ptr_uart->clock/(float)16/(float)baudrate) - (int)(ptr_uart->clock/16/baudrate)) * 32)); // calculate fraction 
  
  brCalc = ptr_uart->clock / (brDivider << 4);              // calculate real truncated baud rate 
  // If inside +/- 2% tolerance, baudrate configured correctly
  if (!(((brCalc * 100) < (baudrate * 102)) && ((brCalc * 100) > (baudrate * 98))))
    return ARM_UART_ERROR_BAUDRATE;
  
  /* Fractional part of baud rate to get closer to desired value */
  ptr_uart->uart->C4 = UART_C4_BRFA(brFract);
  
  /* Clear baud rate registers */
  ptr_uart->uart->BDH &= ~(UART_BDH_SBR_MASK);    
  ptr_uart->uart->BDL = 0;
  
  // set baud rate
  ptr_uart->uart->BDH |= (uint8_t)((brDivider >> 8) & 0xFF);
  ptr_uart->uart->BDL |= (uint8_t)((brDivider & 0xFF));
  
  
  /* Data Bits */
  if (data_bits != 8) return ARM_UART_ERROR_DATA_BITS;
  
  /* Parity */
  
  if (parity == ARM_UART_PARITY_NONE)
    ptr_uart->uart->C1 &= ~(UART_C1_PE_MASK);                                          // disable parity (clear bit1)
  else {
    ptr_uart->uart->C1 |= (1U << UART_C1_PE_SHIFT);                                    // enable parity (set bit1)
    switch (parity) {
      case ARM_UART_PARITY_EVEN: ptr_uart->uart->C1 &= ~(UART_C1_PT_MASK);        break;   // set parity to even (clear bit0)
      case ARM_UART_PARITY_ODD:  ptr_uart->uart->C1 |= (1U << UART_C1_PT_SHIFT);  break;   // set parity to odd  (set bit0)
    default: return ARM_UART_ERROR_PARITY;
    }
  }
  
  
  /*
  
  // Stop Bits
  switch (stop_bits) {
    case UART_STOP_BITS_1:    cr2 = 0;                break;
    case UART_STOP_BITS_2:    cr2 = USART_CR2_STOP_1; break;
    case UART_STOP_BITS_1_5:  cr2 = USART_CR2_STOP;   break;
    default: return UART_ERROR_STOP_BITS;
  }
  */
  // Flow Control
  
  ptr_uart->info->rts_cts_flow = false;
  switch (flow_control) {
    case ARM_UART_FLOW_CONTROL_NONE: 
      ptr_uart->uart->MODEM = 0; 
      break;
    case ARM_UART_FLOW_CONTROL_RTS_CTS:
      if ((ptr_uart->hw_flow == true) && (ptr_uart->manual_rts_cts == false)) {
        ptr_uart->info->rts_cts_flow = true;
        ptr_uart->uart->MODEM = UART_MODEM_TXCTSE_MASK;// USART_CR3_CTSE;
        //if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_CTS))
        //  ptr_uart->uart->CR3 |= USART_CR3_CTSIE;
      } else {
        return ARM_UART_ERROR_FLOW_CONTROL;
      }
      break;
    default: return ARM_UART_ERROR_FLOW_CONTROL;
  }

  if (rts_pin == 0)
   GPIO_PinWrite(ptr_uart->pin_rts.port, ptr_uart->pin_rts.num, 0);
   
  
  /* Enable UART RX and TX */
  ptr_uart->uart->C2 |= ((1U << UART_C2_RE_SHIFT) | (1U << UART_C2_TE_SHIFT));
  
 
  
  return ARM_UART_OK; 
}

	
#if (RTE_UART0 != 0)
  static ARM_UART_STATUS UART0_Configure (uint32_t           baudrate,
                                      uint8_t            data_bits,
                                      UART_PARITY        parity,
                                      UART_STOP_BITS     stop_bits,
                                      UART_FLOW_CONTROL  flow_control) {
    return UART_Configure (baudrate, data_bits, parity,
                           stop_bits, flow_control, &UART0_Resources);
  }
#endif

#if (RTE_UART1 != 0)
  static ARM_UART_STATUS UART1_Configure (uint32_t           baudrate,
                                      uint8_t            data_bits,
                                      ARM_UART_PARITY        parity,
                                      ARM_UART_STOP_BITS     stop_bits,
                                      ARM_UART_FLOW_CONTROL  flow_control) {
    return UART_Configure (baudrate, data_bits, parity,
                           stop_bits, flow_control, &UART1_Resources);
  }
#endif

#if (RTE_UART2 != 0)
  static ARM_UART_STATUS UART2_Configure (uint32_t           baudrate,
                                      uint8_t            data_bits,
                                      ARM_UART_PARITY        parity,
                                      ARM_UART_STOP_BITS     stop_bits,
                                      ARM_UART_FLOW_CONTROL  flow_control) {
    return UART_Configure (baudrate, data_bits, parity,
                           stop_bits, flow_control, &UART2_Resources);
  }
#endif

//static void UART_LON_IRQHandler (UART_RESOURCES *ptr_uart) {
//  if (ptr_uart->uart->S2 & UART_S2_LBKDIF_MASK) {
//    ptr_uart->uart->S2 |= (1U << UART_S2_LBKDIF_SHIFT);
//    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_BREAK)) {
//      ptr_uart->info->cb_event(ARM_UART_EVENT_RX_BREAK);
//    }
//  }
//}

    
    
/**
  \fn          int32_t UART_WriteData (const uint8_t   *data,
                                       uint32_t         size,
                                       UART_RESOURCES  *ptr_uart)
  \brief       Write data to UART transmitter.
  \param[in]   data  Pointer to buffer with data to write to UART transmitter
  \param[in]   size  Data buffer size in bytes
  \param[in]   ptr_uart Pointer to UART resources
  \return      number of data bytes written, error code if negative
*/
static int32_t UART_WriteData (const uint8_t   *data, 
                                     uint32_t   size,
                                     UART_RESOURCES  *ptr_uart) {
  uint32_t cnt, offs;
  int32_t  wr_cnt = 0;

  #if DMA_TX                               
  uint32_t tick;
  #endif
  
  if (ptr_uart->info->power == false) { 
    return wr_cnt; 
  }
  
#if   defined(RTE_CMSIS_RTOS2)
  osMutexAcquire(ptr_uart->info->mutex_write, osWaitForever);
#elif defined(RTE_CMSIS_RTOS)
  osMutexWait(ptr_uart->info->mutex_write, osWaitForever);
#endif

  /* Determine number of data bytes in buffer */
  cnt = (ptr_uart->info->tx.idx_in - ptr_uart->info->tx.idx_out) & (ptr_uart->tx_buffer_size - 1);

  /* Check if free space in the tx buffer */
  if (size && !((cnt == 0) && !ptr_uart->info->tx.empty)) {

    /* Buffer not full - get available space in it */
    cnt = ptr_uart->tx_buffer_size - cnt;

    if (size > cnt) {
      /* Max cnt of data can be written */
      size = cnt;
    }

    offs = ptr_uart->info->tx.idx_in;

    if (size > (ptr_uart->tx_buffer_size - offs)) {
      /* Write data from in to the end of tx buffer */
      wr_cnt = ptr_uart->tx_buffer_size - offs;
      memcpy(&ptr_uart->tx_buffer[offs], data, wr_cnt);
      
      /* Adjust number of data to write and address offset */
      size -= wr_cnt;
      offs  = 0;
    }

    memcpy(&ptr_uart->tx_buffer[offs], &data[wr_cnt], size);
    wr_cnt += size;

    ptr_uart->info->tx.idx_in = (ptr_uart->info->tx.idx_in + wr_cnt) & (ptr_uart->tx_buffer_size - 1);
    ptr_uart->info->tx.empty  = false;
  }
  osMutexRelease(ptr_uart->info->mutex_write);
  
  if (ptr_uart->dma_tx == NULL) {
    ptr_uart->uart->C2 |= (1U << UART_C2_TIE_SHIFT);
  } 
  
  
#if DMA_TX
  else {
    
#if   defined(RTE_CMSIS_RTOS2)
    if (osSemaphoreAcquire(ptr_uart->info->semaphore, 0) > 0) {
#elif defined(RTE_CMSIS_RTOS)
    if (osSemaphoreWait(ptr_uart->info->semaphore, 0) > 0) {
#endif
      DMA_ChannelDisable(ptr_uart->dma_tx->channel);

#if   defined(RTE_CMSIS_RTOS2)
      tick = osKernelGetTickCount();
#elif defined(RTE_CMSIS_RTOS)
      tick = osKernelSysTick();
#endif
      do {
        if (DMA_ChannelStatus(ptr_uart->dma_tx->channel) == false) break;
#if   defined(RTE_CMSIS_RTOS2)
      } while ((osKernelGetTickCount() - tick) < ((osKernelGetTickFreq() + 999U) / 1000U));
#elif defined(RTE_CMSIS_RTOS)
      } while ((osKernelSysTick() - tick) < osKernelSysTickMicroSec(1000));
#endif

      if ((DMA_ChannelStatus(ptr_uart->dma_tx->channel) == false)) {
        DMA_ChannelConfigure(ptr_uart->dma_tx, 
                            (uint32_t)ptr_uart->tx_buffer + ptr_uart->info->tx.idx_out,    // src address
                            (uint32_t)(&ptr_uart->uart->D),                                // dest address
                             wr_cnt,                                                       // data count
                             DMA_DATA_SIZE_8BIT, DMA_DATA_SIZE_8BIT, 1, 0,                 // src data size, dest data size, src offset, dest offset (byte)
                             DMA_MAJORITR_INT_ENABLE);                                     // interrupt enable
        ptr_uart->info->tx.transfer_len = wr_cnt;
        DMA_ChannelEnable (ptr_uart->dma_tx->channel);  
        
          
     }
    }
  } 
#endif     
  

  return wr_cnt;
}

#if (RTE_UART0)
  static int32_t UART0_WriteData (const uint8_t *data, uint32_t size) {
    return UART_WriteData (data, size, &UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static int32_t UART1_WriteData (const uint8_t *data, uint32_t size) {
    return UART_WriteData (data, size, &UART1_Resources);
  }
#endif

#if (RTE_UART2)
  static int32_t UART2_WriteData (const uint8_t *data, uint32_t size) {
    return UART_WriteData (data, size, &UART2_Resources);
  }
#endif



/**
  \fn          int32_t UART_ReadData (uint8_t *data,
                                      uint32_t         size,
                                      UART_RESOURCES  *ptr_uart)
  \brief       Read data from UART receiver.
  \param[out]  data  Pointer to buffer for data read from UART receiver
  \param[in]   size  Data buffer size in bytes
  \param[in]   ptr_uart Pointer to UART resources
  \return      number of data bytes read, error code if negative
*/
static int32_t UART_ReadData (uint8_t         *data,
                              uint32_t         size,
                              UART_RESOURCES  *ptr_uart) {
  uint32_t cnt, offs, out;
  int32_t  rd_cnt = 0;
  
  if (ptr_uart->info->power == false) { return rd_cnt; }

#if   defined(RTE_CMSIS_RTOS2)
  osMutexAcquire(ptr_uart->info->mutex_read, osWaitForever);
#elif defined(RTE_CMSIS_RTOS)
  osMutexWait(ptr_uart->info->mutex_read, osWaitForever);
#endif

  if (size && !ptr_uart->info->rx.empty) {
    /* Determine number of data bytes in buffer */
    cnt = (ptr_uart->info->rx.idx_in - ptr_uart->info->rx.idx_out) & (ptr_uart->rx_buffer_size - 1);

    if (cnt == 0) {
      /* Rx buffer is full */
      cnt = ptr_uart->rx_buffer_size;
    }

    if (size > cnt) {
      /* Max val of data can be read */
      size = cnt;
    }
    
    offs = ptr_uart->info->rx.idx_out;

    if (size > (ptr_uart->rx_buffer_size - offs)) {
      /* Read data from out to the end of rx buffer */
      rd_cnt = ptr_uart->rx_buffer_size - offs;
      memcpy(data, &ptr_uart->rx_buffer[offs], rd_cnt);
      
      /* Adjust number of data to read and address offset */
      size -= rd_cnt;
      offs  = 0;
    }

    memcpy(&data[rd_cnt], &ptr_uart->rx_buffer[offs], size);
    rd_cnt += size;
    
    /* Precalculate output index to determine if buffer empty */
    out = (ptr_uart->info->rx.idx_out + rd_cnt) & (ptr_uart->rx_buffer_size - 1);

    if (out == ptr_uart->info->rx.idx_in) {
        ptr_uart->info->rx.empty = true;
    }
    ptr_uart->info->rx.idx_out = out;
  }

  /* Enable RX Data Valid interrupt */
  /* enable rx interrupt*/ 
  ptr_uart->uart->C2 |= (1U << UART_C2_RIE_SHIFT);

  osMutexRelease(ptr_uart->info->mutex_read);

  return rd_cnt;
}

#if (RTE_UART0)
  static int32_t UART0_ReadData (uint8_t *data, uint32_t size) {
    return UART_ReadData (data, size, &UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static int32_t UART1_ReadData (uint8_t *data, uint32_t size) {
    return UART_ReadData (data, size, &UART1_Resources);
  }
#endif

#if (RTE_UART2)
  static int32_t UART2_ReadData (uint8_t *data, uint32_t size) {
    return UART_ReadData (data, size, &UART2_Resources);
  }
#endif

/**
  \fn          int32_t UART_DataAvailable (UART_RESOURCES  *ptr_uart)
  \brief       Check available data in UART receiver.
  \param[in]   ptr_uart Pointer to UART resources
  \return      number of data bytes in receive buffer, error code if negative
*/
static int32_t UART_DataAvailable (UART_RESOURCES *ptr_uart) {
  uint32_t cnt;

  cnt = (ptr_uart->info->rx.idx_in - ptr_uart->info->rx.idx_out) & (ptr_uart->rx_buffer_size - 1);

  if ((cnt == 0) && !ptr_uart->info->rx.empty) {
    /* Buffer is full */
    cnt = ptr_uart->rx_buffer_size;
  }
  return (cnt);
}

#if (RTE_UART0)
  static int32_t UART0_DataAvailable (void) {
    return UART_DataAvailable (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static int32_t UART1_DataAvailable (void) {
    return UART_DataAvailable (&UART1_Resources);
  }
#endif

#if (RTE_UART2)
  static int32_t UART2_DataAvailable (void) {
    return UART_DataAvailable (&UART2_Resources);
  }
#endif

/**
  \fn          bool UART_TxDone (UART_RESOURCES  *ptr_uart)
  \brief       Check if UART transmission is completed.
  \param[in]   ptr_uart Pointer to UART resources
  \return      transmitter completion status
                 - \b true transmission done
                 - \b false transmission in progress
*/
static bool UART_TxDone (UART_RESOURCES *ptr_uart) {
  return (ptr_uart->info->tx.empty);
}

#if (RTE_UART0)
  static bool UART0_TxDone (void) {
    return UART_TxDone (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static bool UART1_TxDone (void) {
    return UART_TxDone (&UART1_Resources);
  }
#endif

#if (RTE_UART2)
  static bool UART2_TxDone (void) {
    return UART_TxDone (&UART2_Resources);
  }
#endif


/**
  \fn          ARM_UART_STATUS UART_SetTxThreshold (uint32_t         level,
                                                UART_RESOURCES  *ptr_uart) {
  \brief       Set transmit threshold for UART_TX_THRESHOLD event.
  \param[in]   level  Number of character in transmit buffer
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_SetTxThreshold (uint32_t level, UART_RESOURCES *ptr_uart) {

  if (level > ptr_uart->tx_buffer_size) { return ARM_UART_ERROR; }
  ptr_uart->info->tx.threshold = level;

  return ARM_UART_OK;
}

#if (RTE_UART0)
  static ARM_UART_STATUS UART0_SetTxThreshold (uint32_t level) {
    return UART_SetTxThreshold (level, &UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_STATUS UART1_SetTxThreshold (uint32_t level) {
    return UART_SetTxThreshold (level, &UART1_Resources);
  }
#endif

#if (RTE_UART2)
  static ARM_UART_STATUS UART2_SetTxThreshold (uint32_t level) {
    return UART_SetTxThreshold (level, &UART2_Resources);
  }
#endif

/**
  \fn          ARM_UART_STATUS UART_SetRxThreshold (uint32_t         level,
                                                UART_RESOURCES  *ptr_uart) {
  \brief       Set receive threshold for UART_RX_THRESHOLD event.
  \param[in]   level  Number of character in transmit buffer
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_SetRxThreshold (uint32_t level, UART_RESOURCES *ptr_uart) {

  if (level > ptr_uart->rx_buffer_size) { return ARM_UART_ERROR; }
  ptr_uart->info->rx.threshold = level;
  
  return ARM_UART_OK;
}

#if (RTE_UART0)
  static ARM_UART_STATUS UART0_SetRxThreshold (uint32_t level) {
    return UART_SetRxThreshold (level, &UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_STATUS UART1_SetRxThreshold (uint32_t level) {
    return UART_SetRxThreshold (level, &UART1_Resources);
  }
#endif

#if (RTE_UART2)
  static ARM_UART_STATUS UART2_SetRxThreshold (uint32_t level) {
    return UART_SetRxThreshold (level, &UART2_Resources);
  }
#endif

/**
  \fn          ARM_UART_STATUS UART_FlushTxBuffer (UART_RESOURCES  *ptr_uart)
  \brief       Flush UART transmit buffer.
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_FlushTxBuffer (UART_RESOURCES *ptr_uart) {

#if   defined(RTE_CMSIS_RTOS2)
  osMutexAcquire(ptr_uart->info->mutex_write, osWaitForever);
#elif defined(RTE_CMSIS_RTOS)
  osMutexWait(ptr_uart->info->mutex_write, osWaitForever);
#endif
  
  /* Disable transmitter */
  ptr_uart->uart->C2  &= ~UART_C2_TE_MASK;
  /* Disable tx buffer empty interrupt */
  ptr_uart->uart->C2 &= ~UART_C2_TIE_MASK;
  
  /* Clear tx peripheral buffer and shift register */
  //ptr_uart->uart->D  = USART_CMD_CLEARTX;

  /* Reset buffer info */
  ptr_uart->info->tx.idx_in  = 0;
  ptr_uart->info->tx.idx_out = 0;
  ptr_uart->info->tx.empty   = true;

  /* Enable transmitter */
  ptr_uart->uart->C2 |= (1U << UART_C2_TE_SHIFT);

  osMutexRelease(ptr_uart->info->mutex_write);
  
  return ARM_UART_OK;
}

#if (RTE_UART0)
  static ARM_UART_STATUS UART0_FlushTxBuffer (void) {
    return UART_FlushTxBuffer (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_STATUS UART1_FlushTxBuffer (void) {
    return UART_FlushTxBuffer (&UART1_Resources);
  }
#endif

#if (RTE_UART2)
  static ARM_UART_STATUS UART2_FlushTxBuffer (void) {
    return UART_FlushTxBuffer (&UART2_Resources);
  }
#endif

/**
  \fn          ARM_UART_STATUS UART_FlushRxBuffer (UART_RESOURCES  *ptr_uart)
  \brief       Flush UART transmit buffer.
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_FlushRxBuffer (UART_RESOURCES *ptr_uart) {

#if   defined(RTE_CMSIS_RTOS2)
  osMutexAcquire(ptr_uart->info->mutex_read, osWaitForever);
#elif defined(RTE_CMSIS_RTOS)
  osMutexWait(ptr_uart->info->mutex_read, osWaitForever);
#endif

  /* Disable receiver */
  ptr_uart->uart->C2  &= ~UART_C2_RE_MASK;
  /* Disable RX data valid interrupt */
  ptr_uart->uart->C2 &= ~UART_C2_RIE_MASK;
  /* Clear rx peripheral buffer and shift register */
  // ptr_uart->uart->CMD  = USART_CMD_CLEARRX;

  /* Reset buffer info */
  ptr_uart->info->rx.idx_in  = 0;
  ptr_uart->info->rx.idx_out = 0;
  ptr_uart->info->rx.empty   = true;

  /* Enable rx data valid interrupt */
  ptr_uart->uart->C2 |= (1U << UART_C2_RIE_SHIFT);
  /* Enable receiver */
  ptr_uart->uart->C2 |= UART_C2_RE_MASK;

  osMutexRelease(ptr_uart->info->mutex_read);

  return ARM_UART_OK;
}

#if (RTE_UART0)
  static ARM_UART_STATUS UART0_FlushRxBuffer (void) {
    return UART_FlushRxBuffer (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_STATUS UART1_FlushRxBuffer (void) {
    return UART_FlushRxBuffer (&UART1_Resources);
  }
#endif

#if (RTE_UART2)
  static ARM_UART_STATUS UART2_FlushRxBuffer (void) {
    return UART_FlushRxBuffer (&UART2_Resources);
  }
#endif

/**
  \fn          ARM_UART_COM_ERROR UART_GetComError (UART_RESOURCES  *ptr_uart)
  \brief       Get UART Communication errors.
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_COM_ERROR
*/
static ARM_UART_COM_ERROR UART_GetComError (UART_RESOURCES  *ptr_uart) {
  ARM_UART_COM_ERROR com_error = {0, 0, 0, 0};
  return com_error;
}

#if (RTE_UART0)
  static ARM_UART_COM_ERROR UART0_GetComError (void) {
    return UART_GetComError (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_COM_ERROR UART1_GetComError (void) {
    return UART_GetComError (&UART1_Resources);
  }
#endif

#if (RTE_UART2)
  static ARM_UART_COM_ERROR UART2_GetComError (void) {
    return UART_GetComError (&UART2_Resources);
  }
#endif

/**
  \fn          ARM_UART_STATUS UART_SetComBreak (UART_RESOURCES  *ptr_uart)
  \brief       Suspend transmission and put UART Transmission line in a break state.
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_SetComBreak (UART_RESOURCES  *ptr_uart) {
  if (ptr_uart->info->power == false) { return ARM_UART_ERROR; }
  
  ptr_uart->info->com_break = true;
 // ptr_uart->uart->IEN |= USART_IEN_TXBL;
  return ARM_UART_OK;
}

#if (RTE_UART0)
  static ARM_UART_STATUS UART0_SetComBreak (void) {
    return UART_SetComBreak (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_STATUS UART1_SetComBreak (void) {
    return UART_SetComBreak (&UART1_Resources);
  }
#endif

#if (RTE_UART2)
  static ARM_UART_STATUS UART2_SetComBreak (void) {
    return UART_SetComBreak (&UART2_Resources);
  }
#endif

/**
  \fn          ARM_UART_STATUS UART_ClearComBreak (UART_RESOURCES  *ptr_uart)
  \brief       Suspend transmission and put UART Transmission line in a break state.
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_ClearComBreak (UART_RESOURCES  *ptr_uart) {
  if (ptr_uart->info->power == false) { return ARM_UART_ERROR; }

  ptr_uart->info->com_break = false;
  //ptr_uart->uart->IEN |= USART_IEN_TXBL;
  return ARM_UART_OK;
}

#if (RTE_UART0)
  static ARM_UART_STATUS UART0_ClearComBreak (void) {
    return UART_ClearComBreak (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_STATUS UART1_ClearComBreak (void) {
    return UART_ClearComBreak (&UART1_Resources);
  }
#endif
	
#if (RTE_UART2)
  static ARM_UART_STATUS UART2_ClearComBreak (void) {
    return UART_ClearComBreak (&UART2_Resources);
  }
#endif

/**
  \fn          ARM_UART_STATUS UART_SetModemControl (ARM_UART_MODEM_CONTROL control,
                                                 UART_RESOURCES    *ptr_uart)
  \brief       Set UART Modem Control line state.
  \param[in]   control  \ref ARM_UART_MODEM_CONTROL
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_SetModemControl (ARM_UART_MODEM_CONTROL control,
                                         UART_RESOURCES    *ptr_uart) {
  return ARM_UART_ERROR;
}

#if (RTE_UART0)
static ARM_UART_STATUS UART0_SetModemControl (ARM_UART_MODEM_CONTROL control) {
  return UART_SetModemControl (control, &UART0_Resources);
}
#endif

#if (RTE_UART1)
static ARM_UART_STATUS UART1_SetModemControl (ARM_UART_MODEM_CONTROL control) {
  return UART_SetModemControl (control, &UART1_Resources);
}
#endif

#if (RTE_UART2)
static ARM_UART_STATUS UART2_SetModemControl (ARM_UART_MODEM_CONTROL control) {
  return UART_SetModemControl (control, &UART2_Resources);
}
#endif

/**
  \fn          ARM_UART_MODEM_STATUS UART_GetModemStatus (UART_RESOURCES *ptr_uart)
  \brief       Get UART Modem Status lines state.
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_MODEM_STATUS
*/
static ARM_UART_MODEM_STATUS UART_GetModemStatus (UART_RESOURCES *ptr_uart) {
  ARM_UART_MODEM_STATUS mst = {0, 0, 0, 0};
  return mst;
}

#if (RTE_UART0)
static ARM_UART_MODEM_STATUS UART0_GetModemStatus (void) {
  return UART_GetModemStatus (&UART0_Resources);
}
#endif

#if (RTE_UART1)
static ARM_UART_MODEM_STATUS UART1_GetModemStatus (void) {
  return UART_GetModemStatus (&UART1_Resources);
}
#endif

#if (RTE_UART2)
static ARM_UART_MODEM_STATUS UART2_GetModemStatus (void) {
  return UART_GetModemStatus (&UART2_Resources);
}
#endif

static void UART_RX_TX_IRQHandler (UART_RESOURCES *ptr_uart) {

  uint32_t flags1 = ptr_uart->uart->S1;
  uint32_t flags2 = ptr_uart->uart->S2; 
  uint32_t cnt;
  volatile uint8_t dwCount;
  volatile uint8_t tmc;


  if (flags1 & UART_S1_RDRF_MASK) {
    /* Received data valid interrupt */
    cnt = (ptr_uart->info->rx.idx_in - ptr_uart->info->rx.idx_out) & (ptr_uart->rx_buffer_size - 1);
  
    if (cnt || ptr_uart->info->rx.empty) {
      /* Free space in buffer or buffer empty */
      ptr_uart->rx_buffer[ptr_uart->info->rx.idx_in] = ptr_uart->uart->D;

      ptr_uart->info->rx.idx_in += 1;
      ptr_uart->info->rx.idx_in &= ptr_uart->rx_buffer_size - 1;

      if (ptr_uart->info->rx.empty) {
        if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_NOT_EMPTY)) {
          ptr_uart->info->cb_event(ARM_UART_EVENT_RX_NOT_EMPTY);
        }
      }
      ptr_uart->info->rx.empty = false;

      cnt += 1;

      if (cnt == ptr_uart->info->rx.threshold) {
        if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_THRESHOLD)) {
          ptr_uart->info->cb_event(ARM_UART_EVENT_RX_THRESHOLD);
        }
      }
    }
    else {
      /* Rx buffer full, disable data available interrupt */
      ptr_uart->uart->C2 &= ~UART_C2_RIE_MASK;
    }
  }


  else if (flags1 & UART_S1_TDRE_MASK) {
    /* Peripheral tx buffer empty interrupt */
    if (ptr_uart->info->tx.empty) {
      if (ptr_uart->info->com_break) {
        /* Send break condition */
        ptr_uart->uart->C2 |= (1U << UART_C2_SBK_SHIFT); 
      }
      /* Disable TX empty interrupt */
      ptr_uart->uart->C2 &= ~UART_C2_TIE_MASK;
    }
    else {
      /* Data to transmit exists in tx buffer */
      ptr_uart->uart->D = ptr_uart->tx_buffer[ptr_uart->info->tx.idx_out];

      ptr_uart->info->tx.idx_out += 1;
      ptr_uart->info->tx.idx_out &= ptr_uart->tx_buffer_size - 1;

      /* Determine how may data is in the buffer */
      cnt = (ptr_uart->info->tx.idx_in - ptr_uart->info->tx.idx_out) & (ptr_uart->tx_buffer_size - 1);

      if ((ptr_uart->tx_buffer_size - cnt) == ptr_uart->info->tx.threshold) {
        /* Free space in buffer reached the tx threshold */
        if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_TX_THRESHOLD)) {
          ptr_uart->info->cb_event(ARM_UART_EVENT_TX_THRESHOLD);
        }
      }

      if (cnt == 0) {
        /* All data from tx buffer sent - buffer is empty */
        ptr_uart->info->tx.empty = true;

        if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_TX_EMPTY)) {
          ptr_uart->info->cb_event(ARM_UART_EVENT_TX_EMPTY);
        }
      }
    }
  }


  /* Idle line detect */
  else if (flags1 & UART_S1_IDLE_MASK) {
    ptr_uart->uart->D;
  }

  /* Receiver overrun */
  else if  (flags1 & UART_S1_OR_MASK) {
    ptr_uart->uart->D;
    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_OVERRUN)) {
      ptr_uart->info->cb_event(ARM_UART_EVENT_RX_OVERRUN);
    }
  }

  /* Noise detect */
  else if ((flags1 & UART_S1_NF_MASK) && ~(flags2 & UART_S2_LBKDE_MASK)) {   
    ptr_uart->uart->D;
  }

  /* Transmition complete */
  else if (flags1 & UART_S1_TC_MASK){
    ptr_uart->uart->C2 |= (1U << UART_C2_TE_SHIFT);

    if (ptr_uart->info->tx.empty == true) {
      if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_TX_DONE))
        ptr_uart->info->cb_event(ARM_UART_EVENT_TX_DONE);
    }
  }

  /* Transmit data register empty */
  else if (flags1 & UART_S1_TDRE_MASK){
    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_TX_EMPTY))
      ptr_uart->info->cb_event(ARM_UART_EVENT_TX_EMPTY);
  }


  // CTS input change
  
}

static void UART_ERR_IRQHandler (UART_RESOURCES *ptr_uart) {

  uint32_t sr1 = ptr_uart->uart->S1;
 
  /* Parity error */
  if (sr1 & UART_S1_PF_MASK) {
    ptr_uart->uart->D;
    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_PARITY_ERROR))
      ptr_uart->info->cb_event(ARM_UART_EVENT_RX_PARITY_ERROR);  
  }

  /* Framing error */
  else if (sr1 & UART_S1_FE_MASK) {
    ptr_uart->uart->D;
    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_FRAMING_ERROR))
      ptr_uart->info->cb_event(ARM_UART_EVENT_RX_FRAMING_ERROR);
  }


  /* Receiver overrun error */
  else if (sr1 & UART_S1_OR_MASK) {
    ptr_uart->uart->D;
    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_OVERRUN))
      ptr_uart->info->cb_event(ARM_UART_EVENT_RX_OVERRUN);
  }

}


#if (RTE_UART0 != 0)
  void UART0_LON_IRQHandler (void) {
    UART_LON_IRQHandler (&UART0_Resources);
  }

  void UART0_RX_TX_IRQHandler (void) {
    UART_RX_TX_IRQHandler (&UART0_Resources);
  }
  
  void UART0_ERR_IRQHandler (void) {
    UART_ERR_IRQHandler(&UART0_Resources);
  }
#endif
  
#if (RTE_UART1 != 0)
  void UART1_RX_TX_IRQHandler (void) {
    UART_RX_TX_IRQHandler (&UART1_Resources);
  }
  
  void UART1_ERR_IRQHandler (void) {
    UART_ERR_IRQHandler(&UART1_Resources);
  }
#endif
 
#if (RTE_UART2 != 0)
  void UART2_RX_TX_IRQHandler (void) {
    UART_RX_TX_IRQHandler (&UART2_Resources);
  }

  void UART2_ERR_IRQHandler (void) {
    UART_ERR_IRQHandler(&UART2_Resources);
  }
#endif
  
#if DMA_TX  
static void UART_DMA_IRQHandler(uint32_t event, UART_RESOURCES *ptr_uart){
  volatile uint32_t cnt;
  volatile uint32_t val;
           uint32_t tick;

  if (event & DMA_STREAM_TRANSFER_COMPLETE) {
    val = (ptr_uart->info->tx.idx_out + ptr_uart->info->tx.transfer_len) & (ptr_uart->tx_buffer_size - 1);
    if (val == ptr_uart->info->tx.idx_in) {
      ptr_uart->info->tx.empty   = true;
      ptr_uart->info->tx.idx_out = val;
      if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_TX_EMPTY))
        ptr_uart->info->cb_event(ARM_UART_EVENT_TX_EMPTY);
      osSemaphoreRelease(ptr_uart->info->semaphore);
    } else {
      ptr_uart->info->tx.idx_out = val;
      val = (ptr_uart->info->tx.idx_in - ptr_uart->info->tx.idx_out) & (ptr_uart->tx_buffer_size - 1);

      if (ptr_uart->info->tx.threshold != 0) {
        if(val == ptr_uart->info->tx.threshold) {
          if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_TX_THRESHOLD))
            ptr_uart->info->cb_event(ARM_UART_EVENT_TX_THRESHOLD);
        }
        if (val > ptr_uart->info->tx.threshold) val -= ptr_uart->info->tx.threshold;
      }

      cnt = ptr_uart->tx_buffer_size - ptr_uart->info->tx.idx_out;
      if (val > cnt) val = cnt;

      DMA_ChannelDisable(ptr_uart->dma_tx->channel);
      tick = osKernelSysTick();
      do {
        if (DMA_ChannelStatus(ptr_uart->dma_tx->channel) == false) break;
      } while ((osKernelSysTick() - tick) < osKernelSysTickMicroSec(1000));

      if ((DMA_ChannelStatus(ptr_uart->dma_tx->channel) == false)) {
        DMA_ChannelConfigure(ptr_uart->dma_tx, 
                            (uint32_t)ptr_uart->tx_buffer + ptr_uart->info->tx.idx_out,    // src address
                            (uint32_t)(&ptr_uart->uart->D),                                // dest address
                             cnt   ,                                                       // data count
                             DMA_DATA_SIZE_8BIT, DMA_DATA_SIZE_8BIT, 1, 0,                 // src data size, dest data size, src offset, dest offset (byte)
                             DMA_MAJORITR_INT_ENABLE);                                     // interrupt enable
        ptr_uart->info->tx.transfer_len = cnt;
        DMA_ChannelEnable (ptr_uart->dma_tx->channel);            
        ptr_uart->info->tx.transfer_len = val;
        
      }
    }
  }  
}
#endif

#if DMA_TX

  #if (RTE_UART0 != 0)
    void DMA_CHANNELx_EVENT(RTE_UART0_TX_DMA_CHANNEL) (uint32_t event)) {
      UART_DMA_IRQHandler (event, &UART0_Resources);
    }
  #endif
    
  #if (RTE_UART1 != 0)
    void DMA_CHANNELx_EVENT(RTE_UART1_TX_DMA_CHANNEL) (uint32_t event) {
      UART_DMA_IRQHandler (event, &UART1_Resources);
    }
  #endif
   
  #if (RTE_UART2 != 0)
    void DMA_CHANNELx_EVENT(RTE_UART2_TX_DMA_CHANNEL) (uint32_t event) {
      UART_DMA_IRQHandler (event, &UART2_Resources);
    }
  #endif
#endif  
    


#if (RTE_UART0)
  /* UART0 peripheral exports Driver_UART0 */
  ARM_DRIVER_UART Driver_UART0 = {
    UARTX_GetVersion,
    UART0_GetCapabilities,
    UART0_Initialize,
    UART0_Uninitialize,
    UART0_PowerControl,
    UART0_Configure,
    UART0_WriteData,
    UART0_ReadData,
    UART0_DataAvailable,
    UART0_TxDone,
    UART0_SetTxThreshold,
    UART0_SetRxThreshold,
    UART0_FlushTxBuffer,
    UART0_FlushRxBuffer,
    UART0_GetComError,
    UART0_SetComBreak,
    UART0_ClearComBreak,
    UART0_SetModemControl,
    UART0_GetModemStatus
  };
#endif

#if (RTE_UART1)
  /* UART1 peripheral exports Driver_UART1 */
  ARM_DRIVER_UART Driver_UART1 = {
    UARTX_GetVersion,
    UART1_GetCapabilities,
    UART1_Initialize,
    UART1_Uninitialize,
    UART1_PowerControl,
    UART1_Configure,
    UART1_WriteData,
    UART1_ReadData,
    UART1_DataAvailable,
    UART1_TxDone,
    UART1_SetTxThreshold,
    UART1_SetRxThreshold,
    UART1_FlushTxBuffer,
    UART1_FlushRxBuffer,
    UART1_GetComError,
    UART1_SetComBreak,
    UART1_ClearComBreak,
    UART1_SetModemControl,
    UART1_GetModemStatus
  };
#endif

#if (RTE_UART2)
  /* UART2 peripheral exports Driver_UART2 */
  ARM_DRIVER_UART Driver_UART2 = {
    UARTX_GetVersion,
    UART2_GetCapabilities,
    UART2_Initialize,
    UART2_Uninitialize,
    UART2_PowerControl,
    UART2_Configure,
    UART2_WriteData,
    UART2_ReadData,
    UART2_DataAvailable,
    UART2_TxDone,
    UART2_SetTxThreshold,
    UART2_SetRxThreshold,
    UART2_FlushTxBuffer,
    UART2_FlushRxBuffer,
    UART2_GetComError,
    UART2_SetComBreak,
    UART2_ClearComBreak,
    UART2_SetModemControl,
    UART2_GetModemStatus
  };
#endif
  

  
