/* ----------------------------------------------------------------------
 * Copyright (C) 2014-2017 ARM Limited. All rights reserved.
 *  
 * $Date:        19. September 2017
 * $Revision:    V1.2
 *  
 * Driver:       Driver_USBD0
 * Configured:   via RTE_Device.h configuration file 
 * Project:      USB Full/Low-Speed Device Driver for Freescale Kinetis K20
 * ---------------------------------------------------------------------- 
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 * 
 *   Configuration Setting                Value
 *   ---------------------                -----
 *   Connect to hardware via Driver_USBD# = 0
 * -------------------------------------------------------------------- */

/* History:
 *  Version 1.2
 *    Added support for CMSIS-RTOS2
 *  Version 1.1
 *    Repaired Endpoint read
 *  Version 1.0
 *    Initial release
 */
 
#include <stdint.h>
#include <string.h>

#include "RTE_Components.h"

#if       defined(RTE_CMSIS_RTOS2)
#include "cmsis_os2.h"
#elif     defined(RTE_CMSIS_RTOS)
#include "cmsis_os.h"
#endif

#include <MK20D5.h>

#include "rl_usb.h"
#include "Driver_USBD.h"

#include "RTE_Device.h"
#include "RTE_Components.h"

/* USBD Driver ****************************************************************/

#define ARM_USBD_DRV_VERSION ARM_DRV_VERSION_MAJOR_MINOR(1,2) /* USBD driver version */

/* Driver Version */
static const ARM_DRV_VERSION usbd_driver_version = { ARM_USBD_API_VERSION, ARM_USBD_DRV_VERSION };

/* Driver Capabilities */
static const ARM_USBD_CAPABILITIES usbd_driver_capabilities = {
  false,  /* event_power_on      */
  false,  /* event_power_off     */
  false,  /* event_connect       */
  false,  /* event_disconnect    */
  true,   /* event_reset         */
  false,  /* event_high_speed    */
  true,   /* event_suspend       */
  true,   /* event_resume        */
  false,  /* event_remote_wakeup */
  false   /* reserved            */
};


#define USBD_DRV_VERSION DRV_VERSION_MAJOR_MINOR(1,0) /* driver version */

// DEFINITIONS
// Number of Endpoints
#if    (RTE_USB_USB0_DEV_EP & 0x8000)
#define USBD_EP_NUM      15
#elif  (RTE_USB_USB0_DEV_EP & 0x4000)
#define USBD_EP_NUM      14
#elif  (RTE_USB_USB0_DEV_EP & 0x2000)
#define USBD_EP_NUM      13
#elif  (RTE_USB_USB0_DEV_EP & 0x1000)
#define USBD_EP_NUM      12
#elif  (RTE_USB_USB0_DEV_EP & 0x0800)
#define USBD_EP_NUM      11
#elif  (RTE_USB_USB0_DEV_EP & 0x0400)
#define USBD_EP_NUM      10
#elif  (RTE_USB_USB0_DEV_EP & 0x0200)
#define USBD_EP_NUM      9
#elif  (RTE_USB_USB0_DEV_EP & 0x0100)
#define USBD_EP_NUM      8
#elif  (RTE_USB_USB0_DEV_EP & 0x0080)
#define USBD_EP_NUM      7
#elif  (RTE_USB_USB0_DEV_EP & 0x0040)
#define USBD_EP_NUM      6
#elif  (RTE_USB_USB0_DEV_EP & 0x0020)
#define USBD_EP_NUM      5
#elif  (RTE_USB_USB0_DEV_EP & 0x0010)
#define USBD_EP_NUM      4
#elif  (RTE_USB_USB0_DEV_EP & 0x0008)
#define USBD_EP_NUM      3
#elif  (RTE_USB_USB0_DEV_EP & 0x0004)
#define USBD_EP_NUM      2
#elif  (RTE_USB_USB0_DEV_EP & 0x0002)
#define USBD_EP_NUM      1
#else
#define USBD_EP_NUM      0
#endif

#define BD_OWN_MASK        0x80
#define BD_DATA01_MASK     0x40
#define BD_KEEP_MASK       0x20
#define BD_NINC_MASK       0x10
#define BD_DTS_MASK        0x08
#define BD_STALL_MASK      0x04

#define IDX(Ep, dir, Ev_Odd) ((((Ep & 0x0F) * 4) + (2 * dir) + (1 *  Ev_Odd)))

#define SETUP_TOKEN    0x0D
#define IN_TOKEN       0x09
#define OUT_TOKEN      0x01
#define TOK_PID(idx)   ((BD[idx].stat >> 2) & 0x0F)

/* Buffer descriptor table */
typedef struct __BUF_DESC {
  uint8_t    stat;
  uint8_t    reserved;
  uint16_t   bc;
  uint32_t   buf_addr;
} BUF_DESC;

/* Endpoint structure */
typedef struct {
  uint8_t   *buffer;
  uint32_t   bufferIndex;
  uint32_t   dataSize;
  uint32_t   maxPacketSize;
  uint8_t    data01;
  uint8_t    bd_index;
} ENDPOINT;


// GLOBAL VARIABLES
static BUF_DESC __align(512) BD[(USBD_EP_NUM + 1) * 2 * 2];
static uint8_t EPBuf[(USBD_EP_NUM + 1)* 2 * 2][64];

volatile static ENDPOINT InEp[USBD_EP_NUM + 1];
volatile static ENDPOINT OutEp[USBD_EP_NUM + 1];

volatile static bool SetupStage = true;

// Static Variables
static ARM_USBD_SignalDeviceEvent_t   cbDeviceEvent;
static ARM_USBD_SignalEndpointEvent_t cbEndpointEvent;

static ARM_USBD_STATE UsbdState = {0, 0, 0, 0};


/* LOCAL FUNCTIONS */

/**
  \fn          void USBD_Reset (uint8_t ep_addr)
  \brief       Called after usbd reset interrupt to reset configuration
*/
static void USBD_Reset (void) {
  uint32_t i;

  USB0->ISTAT   =  0xFF;                // Clear all interrupt status flags
  USB0->ERRSTAT =  0xFF;                // Clear all error flags
  USB0->ERREN   =  0xFF;                // Enable error interrupt sources
  USB0->ADDR    =  0x00;                // Set default address

  for (i = 0; i < USBD_EP_NUM; i++) {
    USB0->ENDPOINT[i].ENDPT = 0x00;
    BD[InEp[i].bd_index].stat      = 0;
    BD[InEp[i].bd_index + 1].stat  = 0;
    BD[OutEp[i].bd_index].stat     = 0;
    BD[OutEp[i].bd_index + 1].stat = 0;
  }

  USB0->ENDPOINT[0].ENDPT = USB_ENDPT_EPHSHK_MASK | // Enable ep handshaking
                            USB_ENDPT_EPTXEN_MASK | // Enable TX (IN) tran.
                            USB_ENDPT_EPRXEN_MASK;  // Enable RX (OUT) tran.

  OutEp[0].data01 = 0;

  SetupStage = true;

  USB0->CTL    |=  USB_CTL_ODDRST_MASK;
}

/**
  \fn          void SetEndpointWrite (uint8_t ep_addr, uint8_t bd_bank, uint8_t *buf, uint32_t len)
  \brief       Start new IN packet transfer
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \param[in]   bd_bank  specifies buffer descriptor bank index
  \param[out]  buf specifies buffer for data read from Endpoint
  \param[in]   len specifies buffer length
  \return      number of data bytes written
*/
static uint32_t SetEndpointWrite (uint8_t ep_addr, uint8_t bd_bank, uint8_t *buf, uint32_t len) {
  uint8_t num, idx;

  num = ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK;

  idx = InEp[num].bd_index + bd_bank;
  if (BD[idx].stat & BD_OWN_MASK) return 0;

  if (len > InEp[num].maxPacketSize) len = InEp[num].maxPacketSize;

  BD[idx].bc = len;
  memcpy(EPBuf[idx], buf, len);
  InEp[num].data01 ^= 1;

  if (InEp[num].data01) {
    BD[idx].stat  = BD_DTS_MASK;
    BD[idx].stat |= BD_OWN_MASK;
  }
  else {
    BD[idx].stat  = BD_DTS_MASK | BD_DATA01_MASK;
    BD[idx].stat |= BD_OWN_MASK;
  }

  return (len);
}

/**
  \fn          void SetEndpointRead (uint8_t ep_addr, uint8_t bd_bank, uint8_t *buf, uint32_t len)
  \brief       Start new OUT packet transfer
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \param[in]   bd_bank  specifies buffer descriptor bank index
  \param[out]  buf specifies buffer for data read from Endpoint
  \param[in]   len specifies buffer length
  \return      number of data bytes written
*/
static uint32_t SetEndpointRead (uint8_t ep_addr, uint8_t bd_bank, uint8_t *buf, uint32_t len) {
  uint8_t num, idx;

  num = ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK;

  idx = OutEp[num].bd_index + bd_bank;

  if (BD[idx].stat & BD_OWN_MASK) return 0;

  if (len > OutEp[num].maxPacketSize) len = OutEp[num].maxPacketSize;
  BD[idx].bc = len;

  if (OutEp[num].data01) {
    BD[idx].stat  = BD_DTS_MASK | BD_DATA01_MASK;
    BD[idx].stat |= BD_OWN_MASK;
  }
  else {
    BD[idx].stat  = BD_DTS_MASK;
    BD[idx].stat |= BD_OWN_MASK;
  }

  return (len);
}



// USB DEVICE DRIVER FUNCTIONS

/**
  \fn          ARM_DRV_VERSION USBD_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRV_VERSION
*/
static ARM_DRV_VERSION GetVersion (void) { return usbd_driver_version; }

/**
  \fn          ARM_USBD_CAPABILITIES USBD_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_USBD_CAPABILITIES
*/
static ARM_USBD_CAPABILITIES GetCapabilities (void) { return usbd_driver_capabilities; };


/**
  \fn          USBD_STATUS Initialize (USBD_SignalDeviceEvent_t   cb_device_event,
                                       USBD_SignalEndpointEvent_t cb_endpoint_event)
  \brief       Initialize USB Device Interface.
  \param[in]   cb_device_event    Pointer to \ref USBD_SignalDeviceEvent
  \param[in]   cb_endpoint_event  Pointer to \ref USBD_SignalEndpointEvent
  \return      \ref USBD_STATUS
*/
ARM_USBD_STATUS Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                            ARM_USBD_SignalEndpointEvent_t cb_endpoint_event) {

  cbDeviceEvent   = cb_device_event;
  cbEndpointEvent = cb_endpoint_event;

  return ARM_USBD_OK;
}


/**
  \fn          USBD_STATUS Uninitialize (void)
  \brief       De-initialize USB Device Interface.
  \return      \ref USBD_STATUS
*/
ARM_USBD_STATUS Uninitialize (void) {
  return ARM_USBD_OK;
}


/**
  \fn          USBD_STATUS PowerControl (POWER_STATE state)
  \brief       Control USB Device Interface Power.
  \param[in]   state specifies Power state
  \return      \ref USBD_STATUS
*/
ARM_USBD_STATUS PowerControl (ARM_POWER_STATE state) {

  switch (state) {
    case ARM_POWER_OFF:
      NVIC_EnableIRQ(USB0_IRQn);                        // Enable interrupts
      USB0->ISTAT   = 0xFF;                             // Clear interrupt flags

      USB0->CTL     &= ~USB_CTL_USBENSOFEN_MASK;        // disable USB
      USB0->CONTROL &= ~USB_CONTROL_DPPULLUPNONOTG_MASK;// pull down on D+
    
      SIM->SCGC4    &= ~SIM_SCGC4_USBOTG_MASK;          // Disable USBOTG clock

      break;

    case ARM_POWER_LOW:
      return ARM_USBD_ERROR;

    case ARM_POWER_FULL:
      // Enable all clocks needed for USB to function
      // Set USB clock to 48 MHz
      SIM->SOPT2   |=   SIM_SOPT2_USBSRC_MASK     | // MCGPLLCLK used as src
                        SIM_SOPT2_PLLFLLSEL_MASK  ; // Select MCGPLLCLK as clock
      SIM->CLKDIV2 &= ~(SIM_CLKDIV2_USBFRAC_MASK  | // Clear CLKDIV2 FS values
                        SIM_CLKDIV2_USBDIV_MASK);
      SIM->CLKDIV2  =   SIM_CLKDIV2_USBDIV(0)     ; // USB clk = (PLL*1/2)
                                                    //         = ( 48*1/1)=48
      SIM->SCGC4   |=   SIM_SCGC4_USBOTG_MASK;      // Enable USBOTG clock

      USB0->BDTPAGE1 = (uint8_t) ((uint32_t) BD >> 8 );
      USB0->BDTPAGE2 = (uint8_t) ((uint32_t) BD >> 16);
      USB0->BDTPAGE3 = (uint8_t) ((uint32_t) BD >> 24);

      USB0->ISTAT   = 0xFF;                             // Clear interrupt flags

      // Enable interrupts
      USB0->INTEN   =  USB_INTEN_USBRSTEN_MASK |
                       USB_INTEN_TOKDNEEN_MASK |
                       USB_INTEN_SLEEPEN_MASK  |
                       USB_INTEN_ERROREN_MASK;

      USB0->USBCTRL  = USB_USBCTRL_PDE_MASK;            // pull dawn on D+ and D-
      USB0->ISTAT   = 0xFF;                             // Clear interrupt flags
      NVIC_EnableIRQ(USB0_IRQn);                        // Enable interrupts

      break;
    
    default: return ARM_USBD_ERROR;
  }

  return ARM_USBD_OK;
}


/**
  \fn          USBD_STATUS DeviceConnect (void)
  \brief       Connect USB Device.
  \return      \ref USBD_STATUS
*/
ARM_USBD_STATUS DeviceConnect (void) {

  UsbdState.connected = true;
  UsbdState.active    = true;
  
  USB0->CTL    |= USB_CTL_USBENSOFEN_MASK;              // enable USB
  USB0->CONTROL = USB_CONTROL_DPPULLUPNONOTG_MASK;      // pull up on D+

  return ARM_USBD_OK;
}


/**
  \fn          USBD_STATUS DeviceDisconnect (void)
  \brief       Disconnect USB Device.
  \return      \ref USBD_STATUS
*/
ARM_USBD_STATUS DeviceDisconnect (void) {
  USB0->CTL  &= ~USB_CTL_USBENSOFEN_MASK;               // disable USB
  USB0->CONTROL &= ~USB_CONTROL_DPPULLUPNONOTG_MASK;    // pull down on D+
  return ARM_USBD_OK;
}


/**
  \fn          USBD_STATE DeviceGetState (void)
  \brief       Get current USB Device State.
  \return      \ref USBD_STATE
*/
ARM_USBD_STATE DeviceGetState (void) {
  return UsbdState;
}


/**
  \fn          USBD_STATUS DeviceRemoteWakeup (void)
  \brief       Trigger USB Remote Wakeup.
  \return      \ref USBD_STATUS
*/
ARM_USBD_STATUS DeviceRemoteWakeup (void) {
  USB0->CTL |=  USB_CTL_RESUME_MASK;
  osDelay(5);
  USB0->CTL &= ~USB_CTL_RESUME_MASK;

  return ARM_USBD_OK;
}


/**
  \fn          USBD_STATUS DeviceSetAddress (uint8_t dev_addr,
                                             USBD_SET_ADDRESS_STAGE stage)
  \brief       Set USB Device Address.
  \param[in]   dev_addr specifies Device Address
  \param[in]   stage specifies stage in which the function is called
  \return      \ref USBD_STATUS
*/
ARM_USBD_STATUS DeviceSetAddress (uint8_t                    dev_addr,
                                  ARM_USBD_SET_ADDRESS_STAGE stage) {
                                    

  if (stage == ARM_USBD_SET_ADDRESS_STATUS) {
    USB0->ADDR = dev_addr & 0x7F;
  }
  return ARM_USBD_OK;
}


  /**
  \fn          USBD_STATUS DeviceConfigure (bool configure)
  \brief       Configure/unconfigure USB Device.
  \param[in]   configure specifies operation
                - \b false Unconfigure
                - \b true  Configure
  \return      \ref USBD_STATUS
*/
ARM_USBD_STATUS DeviceConfigure (bool configure) {
  return ARM_USBD_OK;
}


/**
  \fn          USBD_STATUS USBD_EndpointConfigure (uint8_t ep_addr,
                                                   USB_ENDPOINT_TYPE ep_type,
                                                   uint16_t ep_max_packet_size)
  \brief       Configure USB Endpoint.
  \param[in]   ptr_epd specifies pointer to Endpoint descriptor
  \return      \ref USBD_STATUS
*/
ARM_USBD_STATUS EndpointConfigure (uint8_t           ep_addr,
                                   ARM_USB_ENDPOINT_TYPE ep_type,
                                   uint16_t          ep_max_packet_size) {
  uint8_t num;

  num = ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK;

  if ((RTE_USB_USB0_DEV_EP & (1 << num)) == 0) return ARM_USBD_ERROR;

  if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) {
    InEp[num].buffer        = NULL;
    InEp[num].dataSize      = 0;
    InEp[num].bufferIndex   = 0;
    InEp[num].data01        = 0;
    InEp[num].bd_index      = num * 4 + 2;
    InEp[num].maxPacketSize = ep_max_packet_size  & ARM_USB_ENDPOINT_MAX_PACKET_SIZE_MASK;

    BD[InEp[num].bd_index].buf_addr     = (uint32_t) &(EPBuf[InEp[num].bd_index][0]);
    BD[InEp[num].bd_index + 1].buf_addr = (uint32_t) &(EPBuf[InEp[num].bd_index + 1][0]);

    BD[InEp[num].bd_index].stat     = 0;
    BD[InEp[num].bd_index + 1].stat = 0;

    InEp[num].data01 = 0;

    USB0->ENDPOINT[num].ENDPT |= USB_ENDPT_EPHSHK_MASK |  // enable ep handshaking
                                 USB_ENDPT_EPTXEN_MASK;   // enable TX (IN) tran.

  } else {
    OutEp[num].buffer         = NULL;
    OutEp[num].dataSize       = 0;
    OutEp[num].bufferIndex    = 0;
    OutEp[num].data01         = 0;
    OutEp[num].bd_index       = num * 4;
    OutEp[num].maxPacketSize  = ep_max_packet_size  & ARM_USB_ENDPOINT_MAX_PACKET_SIZE_MASK;

    BD[OutEp[num].bd_index].buf_addr     = (uint32_t) &(EPBuf[OutEp[num].bd_index][0]);
    BD[OutEp[num].bd_index + 1].buf_addr = (uint32_t) &(EPBuf[OutEp[num].bd_index + 1][0]);

    BD[OutEp[num].bd_index].stat     = 0;
    BD[OutEp[num].bd_index + 1].stat = 0;

    USB0->ENDPOINT[num].ENDPT |= USB_ENDPT_EPHSHK_MASK | // Enable EP handshaking
                                 USB_ENDPT_EPRXEN_MASK;  // Enable RX (OUT) trans.
  }

  return ARM_USBD_OK;
}


/**
  \fn          USBD_STATUS EndpointUnconfigure (uint8_t ep_addr)
  \brief       Unconfigure USB Endpoint.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \return      \ref USBD_STATUS
*/
ARM_USBD_STATUS EndpointUnconfigure (uint8_t ep_addr) {
  uint8_t num;
  volatile ENDPOINT * ep;

  num = ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK;
  if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) {
    ep = & (InEp[num]);
  } else {
    ep = & (OutEp[num]);
  }

  ep->buffer        = NULL;
  ep->dataSize      = 0;
  ep->bufferIndex   = 0;
  ep->data01        = 0;
  ep->bd_index      = 0;
  ep->maxPacketSize = 0;

  BD[ep->bd_index].stat     = 0;
  BD[ep->bd_index + 1].stat = 0;

  if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK)
    USB0->ENDPOINT[num].ENDPT &= ~USB_ENDPT_EPTXEN_MASK;    // Disable TX (IN) tran.
  else
    USB0->ENDPOINT[num].ENDPT &= ~USB_ENDPT_EPRXEN_MASK;    // Disable RX (OUT) trans.
  
  if (USB0->ENDPOINT[num].ENDPT & (USB_ENDPT_EPTXEN_MASK | USB_ENDPT_EPRXEN_MASK))
    USB0->ENDPOINT[num].ENDPT &= ~USB_ENDPT_EPHSHK_MASK;    // Disable EP handshaking

  return ARM_USBD_OK;
}


/**
  \fn          USBD_STATUS EndpointStall (uint8_t ep_addr, bool stall)
  \brief       Set/Clear Stall for USB Endpoint.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \param[in]   stall specifies operation
                - \b false Clear
                - \b true Set
  \return      \ref USBD_STATUS
*/
ARM_USBD_STATUS EndpointStall (uint8_t ep_addr, bool stall) {
  uint32_t  num;
  volatile ENDPOINT * ep;

  num = ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK;
  if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) {
    ep = & (InEp[num]);
  } else {
    ep = & (OutEp[num]);
  }

  if (stall == true) {
    USB0->ENDPOINT[num].ENDPT |= USB_ENDPT_EPSTALL_MASK;
  } else {
    InEp[num].data01  = 0;
    OutEp[num].data01  = 0;

    BD[ep->bd_index].stat     = 0;
    BD[ep->bd_index + 1].stat = 0;

    USB0->ENDPOINT[num].ENDPT &= ~USB_ENDPT_EPSTALL_MASK;

    BD[ep->bd_index].stat     = 0;
    BD[ep->bd_index + 1].stat = 0;

  }
  return ARM_USBD_OK;
}


/**
\fn          USBD_STATUS EndpointReadStart (uint8_t ep_addr, uint8_t *buf, uint32_t len)
\brief       Start USB Endpoint Read operation.
\param[in]   ep_addr specifies Endpoint Address
              ep_addr.0..3: Address
              ep_addr.7:    Direction
\param[out]  buf specifies buffer for data read from Endpoint
\param[in]   len specifies buffer length
\return      \ref USBD_STATUS
*/
ARM_USBD_STATUS EndpointReadStart (uint8_t ep_addr, uint8_t *buf, uint32_t len) {
  uint8_t num, l;

  num = ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK;

  if ((USB0->ENDPOINT[num].ENDPT & USB_ENDPT_EPHSHK_MASK) == 0) 
    return ARM_USBD_ERROR;
  if ((USB0->ENDPOINT[num].ENDPT & USB_ENDPT_EPRXEN_MASK) == 0) 
    return ARM_USBD_ERROR;
  if (BD[OutEp[num].bd_index].stat     & BD_OWN_MASK) 
    return ARM_USBD_ERROR;
  if (BD[OutEp[num].bd_index + 1].stat & BD_OWN_MASK) 
    return ARM_USBD_ERROR;

  // Save buffer information
  OutEp[num].dataSize    =             len;
  OutEp[num].buffer      = (uint8_t *) buf;

  // Set endpoint read start
  l = SetEndpointRead(ep_addr, 0, (uint8_t *)buf, len);

  len = OutEp[num].dataSize - l;
  if (len) {
    SetEndpointRead(ep_addr, 1, (uint8_t *)buf, len);
  }

  if ((num == 0) && (SetupStage == true)) {
    SetupStage = false;
    USB0->CTL &= ~USB_CTL_TXSUSPENDTOKENBUSY_MASK;
  }

  return ARM_USBD_OK;
}


/**
  \fn          int32_t EndpointRead (uint8_t ep_addr, uint8_t *buf, uint32_t len)
  \brief       Read data from USB Endpoint.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \param[out]  buf specifies buffer for data read from Endpoint
  \param[in]   len specifies buffer length
  \return      number of data bytes read, error code if negative
*/
int32_t EndpointRead (uint8_t ep_addr, uint8_t *buf, uint32_t len) {

  len = OutEp[ep_addr].dataSize;

  OutEp[ep_addr].dataSize    = 0;
  OutEp[ep_addr].bufferIndex = 0;
  OutEp[ep_addr].buffer      = NULL;

  return (len);
}


/**
  \fn          int32_t EndpointWrite (uint8_t ep_addr, const uint8_t *buf, uint32_t len)
  \brief       Write data to USB Endpoint.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \param[in]   buf specifies buffer with data to write to Endpoint
  \param[in]   len specifies buffer length
  \return      number of data bytes written, error code if negative
*/
int32_t EndpointWrite (uint8_t ep_addr, const uint8_t *buf, uint32_t len) {
  uint8_t num;

  num = ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK;

  if ((USB0->ENDPOINT[num].ENDPT & USB_ENDPT_EPHSHK_MASK) == 0) return 0;
  if ((USB0->ENDPOINT[num].ENDPT & USB_ENDPT_EPTXEN_MASK) == 0) return 0;
  if (BD[InEp[num].bd_index].stat     & BD_OWN_MASK)            return 0;
  if (BD[InEp[num].bd_index + 1].stat & BD_OWN_MASK)            return 0;

  // Save buffer information
  InEp[num].dataSize    =             len;
  InEp[num].buffer      = (uint8_t *) buf;

  // Set endpoint write
  InEp[num].bufferIndex = SetEndpointWrite(ep_addr, 0, (uint8_t *)buf, len);

  len = InEp[num].dataSize - InEp[num].bufferIndex;
  if (len) {
    InEp[num].bufferIndex += SetEndpointWrite(ep_addr, 1, (uint8_t *)buf, len);
  }

  if ((num == 0) && (SetupStage == true)) {
    SetupStage = false;
    USB0->CTL &= ~USB_CTL_TXSUSPENDTOKENBUSY_MASK;
  }

  return (InEp[num].dataSize);
}


/**
  \fn          USBD_STATUS EndpointAbort (uint8_t ep_addr)
  \brief       Abort current USB Endpoint transfer.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \return      \ref USBD_STATUS
*/
ARM_USBD_STATUS EndpointAbort (uint8_t ep_addr) {
  uint32_t  num;
  volatile ENDPOINT *ep;

  num = ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK;
  if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) {
    ep = & (InEp[num]);
  } else {
    ep = & (OutEp[num]);
  }

  ep->bufferIndex = 0;
  ep->dataSize    = 0;
  ep->data01      = 0;
  ep->buffer      = NULL;

  return ARM_USBD_OK;
}


/**
  \fn          uint16_t GetFrameNumber (void)
  \brief       Get current USB Frame Number.
  \return      Frame Number
*/
uint16_t GetFrameNumber (void) {
  return((USB0->FRMNUML | (USB0->FRMNUMH << 8)) & 0x07FF);
}


// USB Device interrupt handler
/**
  \fn          void USB0_IRQHandler (void)
  \brief       USB0 interrupt handler
*/
void USB0_IRQHandler(void) {
  uint32_t istat, bd_bank, dir, len;
  uint8_t num;

  istat = USB0->ISTAT & USB0->INTEN;

// Reset interrupt
  if (istat & USB_ISTAT_USBRST_MASK) {
    if (UsbdState.connected == true) {
      UsbdState.active = false;
      USBD_Reset();
      cbDeviceEvent(ARM_USBD_EVENT_RESET);
    }
    USB0->ISTAT = USB_ISTAT_USBRST_MASK;
  }

// Suspend interrupt
  if (istat & USB_ISTAT_SLEEP_MASK) {
//    UsbdState.active = false;
    cbDeviceEvent(ARM_USBD_EVENT_SUSPEND);

//    USB0->INTEN = USB_ISTAT_RESUME_MASK;
    USB0->ISTAT = USB_ISTAT_SLEEP_MASK | USB_ISTAT_RESUME_MASK;
  }

// Resume interrupt
  if (istat & USB_ISTAT_RESUME_MASK) {
    UsbdState.active = true;
    cbDeviceEvent(ARM_USBD_EVENT_RESUME);

    USB0->INTEN &= ~USB_ISTAT_RESUME_MASK;
    USB0->ISTAT  =  USB_ISTAT_RESUME_MASK;
  }

// Error interrupt
  if (istat == USB_ISTAT_ERROR_MASK) {
    USB0->ERRSTAT = 0xFF;
    USB0->ISTAT = USB_ISTAT_ERROR_MASK;
  }


// Token interrupt
  if (istat & USB_ISTAT_TOKDNE_MASK) {

    num     = (USB0->STAT >> 4) & 0x0F;
    dir     = (USB0->STAT >> 3) & 0x01;
    bd_bank = (USB0->STAT >> 2) & 0x01;
    
    USB0->ISTAT = USB_ISTAT_TOKDNE_MASK;

/* setup packet                                                               */
    if ((num == 0) && (TOK_PID((IDX(num, dir, bd_bank))) == SETUP_TOKEN)) {
      BD[InEp[0].bd_index].stat     = 0;
      BD[InEp[0].bd_index + 1].stat = 0;
      InEp[0].data01   = 1;
      OutEp[0].data01  = 0;
      
      SetupStage = true;

      if (OutEp[num].data01 == ((BD[OutEp[num].bd_index + bd_bank].stat >> 6) & 1)) {
        OutEp[num].data01 ^= 1;
        OutEp[num].dataSize = BD[OutEp[num].bd_index + bd_bank].bc;
        memcpy(OutEp[num].buffer, EPBuf[OutEp[num].bd_index + bd_bank], OutEp[num].dataSize);
        cbEndpointEvent(num, ARM_USBD_EP_EVENT_SETUP);
      }
    } else {

/* OUT packet                                                                 */
      if (TOK_PID((IDX(num, dir, bd_bank))) == OUT_TOKEN){
        if (OutEp[num].data01 == ((BD[OutEp[num].bd_index + bd_bank].stat >> 6) & 1)) {
          OutEp[num].data01 ^= 1;
          len = BD[OutEp[num].bd_index + bd_bank].bc;
        } else {
          len = 0;
        }
        memcpy(OutEp[num].buffer + OutEp[num].bufferIndex, EPBuf[OutEp[num].bd_index + bd_bank], len);
        OutEp[num].bufferIndex += len;
        OutEp[num].dataSize    -= len;

        // If all OUT data is received (required size or data terminated with ZLP or short packet),
        // set ARM_USBD_EP_EVENT_OUT
        if ((len % OutEp[num].maxPacketSize) || (OutEp[num].dataSize == 0)){
          BD[OutEp[num].bd_index].stat     = 0;
          BD[OutEp[num].bd_index + 1].stat = 0;
          OutEp[num].dataSize = OutEp[num].bufferIndex;

          cbEndpointEvent(num, ARM_USBD_EP_EVENT_OUT);

        // else prepare next out transaction
        } else {
          SetEndpointRead (num, bd_bank, (uint8_t *) (OutEp[num].buffer + OutEp[num].bufferIndex), OutEp[num].dataSize);
        }
      }

/* IN packet                                                                  */
      if (TOK_PID((IDX(num, dir, bd_bank))) == IN_TOKEN) {

        // If all required In data is send, set ARM_USBD_EP_EVENT_IN
        if (InEp[num].dataSize == InEp[num].bufferIndex) {
          if (((BD[InEp[num].bd_index].stat     & BD_OWN_MASK) == 0) &&
              ((BD[InEp[num].bd_index + 1].stat & BD_OWN_MASK) == 0)) {
            InEp[num].bufferIndex = 0;

            cbEndpointEvent(num | 0x80, ARM_USBD_EP_EVENT_IN);
          }

        // else prepare next IN transaction
        } else {
          len = InEp[num].dataSize - InEp[num].bufferIndex;
          len = SetEndpointWrite ( num | ARM_USB_ENDPOINT_NUMBER_MASK, 
                                   bd_bank,
                                  (uint8_t *) (InEp[num].buffer + InEp[num].bufferIndex), 
                                   len);
          InEp[num].bufferIndex += len;
        }
      }
    }
  }
}



ARM_DRIVER_USBD Driver_USBD0 = {
  GetVersion,
  GetCapabilities,
  Initialize,
  Uninitialize,
  PowerControl,
  DeviceConnect,
  DeviceDisconnect,
  DeviceGetState,
  DeviceRemoteWakeup,
  DeviceSetAddress,
  DeviceConfigure,
  EndpointConfigure,
  EndpointUnconfigure,
  EndpointStall,
  EndpointReadStart,
  EndpointRead,
  EndpointWrite,
  EndpointAbort,
  GetFrameNumber
};
